@@ -18,6 +18,7 @@
 #include "src/base/macros.h"
 #include "src/base/platform/time.h"
 #include "src/execution/isolate-inl.h"
+#include "src/execution/v8threads.h"
 #include "src/flags/flags.h"
 #include "src/handles/handles.h"
 #include "src/handles/traced-handles.h"
@@ -166,13 +167,14 @@ class CppgcPlatformAdapter final : public cppgc::Platform {
     return platform_->MonotonicallyIncreasingTime();
   }
 
-  std::shared_ptr<TaskRunner> GetForegroundTaskRunner() final {
+  std::shared_ptr<TaskRunner> GetForegroundTaskRunner(
+      TaskPriority priority) final {
     // If no Isolate has been set, there's no task runner to leverage for
     // foreground tasks. In detached mode the original platform handles the
     // task runner retrieval.
     if (!isolate_ && !is_in_detached_mode_) return nullptr;
 
-    return platform_->GetForegroundTaskRunner(isolate_);
+    return platform_->GetForegroundTaskRunner(isolate_, priority);
   }
 
   std::unique_ptr<JobHandle> PostJob(TaskPriority priority,
@@ -226,8 +228,16 @@ UnifiedHeapConcurrentMarker::CreateConcurrentMarkingVisitor(
 
 void FatalOutOfMemoryHandlerImpl(const std::string& reason,
                                  const SourceLocation&, HeapBase* heap) {
-  V8::FatalProcessOutOfMemory(
-      static_cast<v8::internal::CppHeap*>(heap)->isolate(), reason.c_str());
+  auto* cpp_heap = static_cast<v8::internal::CppHeap*>(heap);
+  auto* isolate = cpp_heap->isolate();
+  DCHECK_NOT_NULL(isolate);
+  if (v8_flags.heap_snapshot_on_oom) {
+    cppgc::internal::ClassNameAsHeapObjectNameScope names_scope(
+        cpp_heap->AsBase());
+    isolate->heap()->heap_profiler()->WriteSnapshotToDiskAfterGC(
+        v8::HeapProfiler::HeapSnapshotMode::kExposeInternals);
+  }
+  V8::FatalProcessOutOfMemory(isolate, reason.c_str());
 }
 
 void GlobalFatalOutOfMemoryHandlerImpl(const std::string& reason,
@@ -266,6 +276,8 @@ class UnifiedHeapConservativeMarkingVisitor final
 class UnifiedHeapMarker final : public cppgc::internal::MarkerBase {
  public:
   UnifiedHeapMarker(Heap* v8_heap, cppgc::internal::HeapBase& cpp_heap,
+                    std::shared_ptr<::heap::base::IncrementalMarkingSchedule>
+                        incremental_schedule,
                     cppgc::Platform* platform,
                     cppgc::internal::MarkingConfig config);
 
@@ -277,7 +289,7 @@ class UnifiedHeapMarker final : public cppgc::internal::MarkerBase {
 
   cppgc::internal::MutatorMarkingState& GetMutatorMarkingState() {
     return static_cast<cppgc::internal::MutatorMarkingState&>(
-        marking_visitor_->marking_state_);
+        marking_visitor_.marking_state_);
   }
 
   UnifiedHeapMarkingState& GetMutatorUnifiedHeapMarkingState() {
@@ -288,32 +300,64 @@ class UnifiedHeapMarker final : public cppgc::internal::MarkerBase {
     return conservative_marking_visitor_;
   }
 
+  void ScheduleIncrementalMarkingTask() final {
+    if (v8_flags.incremental_marking_unified_schedule) {
+      // Not using stand-alone task from CppHeap.
+      return;
+    }
+    MarkerBase::ScheduleIncrementalMarkingTask();
+  }
+
+  void AdvanceMarkingOnAllocationImpl() final {
+    if (v8_flags.incremental_marking_unified_schedule) {
+      mutator_unified_heap_marking_state_.heap()
+          ->incremental_marking()
+          ->AdvanceOnAllocation();
+      return;
+    }
+    MarkerBase::AdvanceMarkingOnAllocationImpl();
+  }
+
  protected:
-  cppgc::Visitor& visitor() final { return *marking_visitor_; }
+  cppgc::Visitor& visitor() final { return marking_visitor_; }
+
   ::heap::base::StackVisitor& stack_visitor() final {
     return conservative_marking_visitor_;
   }
 
+  cppgc::internal::ConcurrentMarkerBase& concurrent_marker() final {
+    return concurrent_marker_;
+  }
+
+  ::heap::base::IncrementalMarkingSchedule& schedule() final {
+    return *schedule_.get();
+  }
+
  private:
   UnifiedHeapMarkingState mutator_unified_heap_marking_state_;
-  std::unique_ptr<MutatorUnifiedHeapMarkingVisitor> marking_visitor_;
+  MutatorUnifiedHeapMarkingVisitor marking_visitor_;
   UnifiedHeapConservativeMarkingVisitor conservative_marking_visitor_;
+  std::shared_ptr<::heap::base::IncrementalMarkingSchedule> schedule_;
+  UnifiedHeapConcurrentMarker concurrent_marker_;
 };
 
-UnifiedHeapMarker::UnifiedHeapMarker(Heap* v8_heap,
-                                     cppgc::internal::HeapBase& heap,
-                                     cppgc::Platform* platform,
-                                     cppgc::internal::MarkingConfig config)
+UnifiedHeapMarker::UnifiedHeapMarker(
+    Heap* v8_heap, cppgc::internal::HeapBase& heap,
+    std::shared_ptr<::heap::base::IncrementalMarkingSchedule>
+        incremental_schedule,
+    cppgc::Platform* platform, cppgc::internal::MarkingConfig config)
     : cppgc::internal::MarkerBase(heap, platform, config),
       mutator_unified_heap_marking_state_(v8_heap, nullptr,
                                           config.collection_type),
-      marking_visitor_(std::make_unique<MutatorUnifiedHeapMarkingVisitor>(
-          heap, mutator_marking_state_, mutator_unified_heap_marking_state_)),
+      marking_visitor_(heap, mutator_marking_state_,
+                       mutator_unified_heap_marking_state_),
       conservative_marking_visitor_(heap, mutator_marking_state_,
-                                    *marking_visitor_) {
-  concurrent_marker_ = std::make_unique<UnifiedHeapConcurrentMarker>(
-      heap_, v8_heap, marking_worklists_, *schedule_, platform_,
-      mutator_unified_heap_marking_state_, config.collection_type);
+                                    marking_visitor_),
+      schedule_(std::move(incremental_schedule)),
+      concurrent_marker_(heap_, v8_heap, marking_worklists_, schedule(),
+                         platform_, mutator_unified_heap_marking_state_,
+                         config.collection_type) {
+  DCHECK_NOT_NULL(schedule_);
 }
 
 void CppHeap::MetricRecorderAdapter::AddMainThreadEvent(
@@ -469,12 +513,14 @@ CppHeap::CppHeap(
 }
 
 CppHeap::~CppHeap() {
-  if (isolate_) {
-    isolate_->heap()->DetachCppHeap();
-  }
+  Terminate();
 }
 
 void CppHeap::Terminate() {
+  // TODO(ahaas): Remove `already_terminated_` once the V8 API
+  // CppHeap::Terminate has been removed.
+  if (already_terminated_) return;
+  already_terminated_ = true;
   // Must not be attached to a heap when invoking termination GCs.
   CHECK(!isolate_);
   // Gracefully terminate the C++ heap invoking destructors.
@@ -553,13 +599,17 @@ class MoveListenerImpl final : public HeapProfilerNativeMoveListener,
 }  // namespace
 
 void CppHeap::AttachIsolate(Isolate* isolate) {
+  // Since a new isolate is attached, we are also allowed to detach it again.
+  is_detached_ = false;
   CHECK(!in_detached_testing_mode_);
   CHECK_NULL(isolate_);
   isolate_ = isolate;
   heap_ = isolate->heap();
+  stack_->SetScanSimulatorCallback(
+      Isolate::IterateRegistersAndStackOfSimulator);
   static_cast<CppgcPlatformAdapter*>(platform())
       ->SetIsolate(reinterpret_cast<v8::Isolate*>(isolate_));
-  if (auto* heap_profiler = isolate_->heap_profiler()) {
+  if (auto* heap_profiler = heap()->heap_profiler()) {
     heap_profiler->AddBuildEmbedderGraphCallback(&CppGraphBuilder::Run, this);
     heap_profiler->set_native_move_listener(
         std::make_unique<MoveListenerImpl>(heap_profiler, this));
@@ -584,21 +634,29 @@ void CppHeap::AttachIsolate(Isolate* isolate) {
   }
 }
 
-void CppHeap::DetachIsolate() {
+void CppHeap::StartDetachingIsolate() {
+  DCHECK(!is_detached_);
+  is_detached_ = true;
   // TODO(chromium:1056170): Investigate whether this can be enforced with a
   // CHECK across all relevant embedders and setups.
   if (!isolate_) return;
 
+  // Store the last thread that owned the isolate, as it is the thread CppHeap
+  // should also get terminated with.
+  heap_thread_id_ = v8::base::OS::GetCurrentThreadId();
+
   // Finish any ongoing garbage collection.
   if (isolate_->heap()->incremental_marking()->IsMarking()) {
     isolate_->heap()->FinalizeIncrementalMarkingAtomically(
         i::GarbageCollectionReason::kExternalFinalize);
   }
   sweeper_.FinishIfRunning();
+}
 
+void CppHeap::DetachIsolate() {
   sweeping_on_mutator_thread_observer_.reset();
 
-  if (auto* heap_profiler = isolate_->heap_profiler()) {
+  if (auto* heap_profiler = heap()->heap_profiler()) {
     heap_profiler->RemoveBuildEmbedderGraphCallback(&CppGraphBuilder::Run,
                                                     this);
     heap_profiler->set_native_move_listener(nullptr);
@@ -613,7 +671,6 @@ void CppHeap::DetachIsolate() {
     detached_override_stack_state_ = heap_->overridden_stack_state();
     override_stack_state_scope_.reset();
   }
-
   isolate_ = nullptr;
   heap_ = nullptr;
   // Any future garbage collections will ignore the V8->C++ references.
@@ -691,8 +748,10 @@ void CppHeap::UpdateGCCapabilitiesFromFlags() {
       v8_flags.decommit_pooled_pages);
 }
 
-void CppHeap::InitializeMarking(CollectionType collection_type,
-                                GarbageCollectionFlags gc_flags) {
+void CppHeap::InitializeMarking(
+    CollectionType collection_type,
+    std::shared_ptr<::heap::base::IncrementalMarkingSchedule> schedule,
+    GarbageCollectionFlags gc_flags) {
   DCHECK(!collection_type_);
 
   if (collection_type == CollectionType::kMinor) {
@@ -735,15 +794,10 @@ void CppHeap::InitializeMarking(CollectionType collection_type,
   current_gc_flags_ = gc_flags;
 
   const cppgc::internal::MarkingConfig marking_config{
-      *collection_type_,
-      StackState::kNoHeapPointers,
-      SelectMarkingType(),
+      *collection_type_, StackState::kNoHeapPointers, SelectMarkingType(),
       IsForceGC(current_gc_flags_)
           ? cppgc::internal::MarkingConfig::IsForcedGC::kForced
-          : cppgc::internal::MarkingConfig::IsForcedGC::kNotForced,
-      v8::base::TimeDelta::FromMilliseconds(
-          v8_flags.incremental_marking_task_delay_ms),
-      v8_flags.incremental_marking_bailout_when_ahead_of_schedule};
+          : cppgc::internal::MarkingConfig::IsForcedGC::kNotForced};
   DCHECK_IMPLIES(!isolate_,
                  (MarkingType::kAtomic == marking_config.marking_type) ||
                      force_incremental_marking_for_testing_);
@@ -753,9 +807,15 @@ void CppHeap::InitializeMarking(CollectionType collection_type,
     compactor_.InitializeIfShouldCompact(marking_config.marking_type,
                                          marking_config.stack_state);
   }
+
+  if (!schedule) {
+    schedule =
+        ::heap::base::IncrementalMarkingSchedule::Create(v8_flags.predictable);
+  }
+
   marker_ = std::make_unique<UnifiedHeapMarker>(
-      isolate_ ? isolate()->heap() : nullptr, AsBase(), platform_.get(),
-      marking_config);
+      isolate_ ? isolate()->heap() : nullptr, AsBase(), schedule,
+      platform_.get(), marking_config);
 }
 
 namespace {
@@ -784,36 +844,39 @@ void CppHeap::StartMarking() {
   marking_done_ = false;
 }
 
-bool CppHeap::AdvanceTracing(v8::base::TimeDelta max_duration) {
-  if (!TracingInitialized()) return true;
+size_t CppHeap::last_bytes_marked() const {
+  DCHECK(TracingInitialized());
+  return TracingInitialized() ? marker_->last_bytes_marked() : 0;
+}
+
+bool CppHeap::AdvanceMarking(v8::base::TimeDelta max_duration,
+                             size_t marked_bytes_limit) {
+  if (!TracingInitialized()) {
+    return true;
+  }
   is_in_v8_marking_step_ = true;
   cppgc::internal::StatsCollector::EnabledScope stats_scope(
       stats_collector(),
       in_atomic_pause_ ? cppgc::internal::StatsCollector::kAtomicMark
                        : cppgc::internal::StatsCollector::kIncrementalMark);
-  const v8::base::TimeDelta deadline =
-      in_atomic_pause_ ? v8::base::TimeDelta::Max() : max_duration;
-  const size_t marked_bytes_limit = in_atomic_pause_ ? SIZE_MAX : 0;
   DCHECK_NOT_NULL(marker_);
   if (in_atomic_pause_) {
     marker_->NotifyConcurrentMarkingOfWorkIfNeeded(
         cppgc::TaskPriority::kUserBlocking);
   }
-  // TODO(chromium:1056170): Replace when unified heap transitions to
-  // bytes-based deadline.
   marking_done_ =
-      marker_->AdvanceMarkingWithLimits(deadline, marked_bytes_limit);
+      marker_->AdvanceMarkingWithLimits(max_duration, marked_bytes_limit);
   DCHECK_IMPLIES(in_atomic_pause_, marking_done_);
   is_in_v8_marking_step_ = false;
   return marking_done_;
 }
 
-bool CppHeap::IsTracingDone() const {
+bool CppHeap::IsMarkingDone() const {
   return !TracingInitialized() || marking_done_;
 }
 
 bool CppHeap::ShouldFinalizeIncrementalMarking() const {
-  return !incremental_marking_supported() || IsTracingDone();
+  return !incremental_marking_supported() || IsMarkingDone();
 }
 
 void CppHeap::EnterProcessGlobalAtomicPause() {
@@ -855,26 +918,31 @@ void CppHeap::ReEnableConcurrentMarking() {
 }
 
 void CppHeap::WriteBarrier(void* object) {
-  isolate()
-      ->heap()
-      ->mark_compact_collector()
-      ->local_marking_worklists()
-      ->cpp_marking_state()
-      ->MarkAndPush(object);
+  auto& header = cppgc::internal::HeapObjectHeader::FromObject(object);
+  marker_->WriteBarrierForObject<
+      cppgc::internal::MarkerBase::WriteBarrierType::kDijkstra>(header);
 }
 
 namespace {
 
-void RecordEmbedderSpeed(GCTracer* tracer, base::TimeDelta marking_time,
-                         size_t marked_bytes) {
-  constexpr auto kMinReportingTime = base::TimeDelta::FromMillisecondsD(0.5);
-  if (marking_time > kMinReportingTime) {
-    tracer->RecordEmbedderSpeed(marked_bytes, marking_time.InMillisecondsF());
-  }
+void RecordEmbedderMarkingSpeed(GCTracer* tracer, base::TimeDelta marking_time,
+                                size_t marked_bytes) {
+  tracer->RecordEmbedderMarkingSpeed(marked_bytes, marking_time);
 }
 
 }  // namespace
 
+void CppHeap::ProcessCrossThreadWeakness() {
+  CHECK(in_atomic_pause_);
+  CHECK(marking_done_);
+
+  if (!TracingInitialized()) {
+    return;
+  }
+
+  marker_->ProcessCrossThreadWeaknessIfNeeded();
+}
+
 void CppHeap::FinishMarkingAndProcessWeakness() {
   CHECK(in_atomic_pause_);
   CHECK(marking_done_);
@@ -906,8 +974,8 @@ void CppHeap::FinishMarkingAndProcessWeakness() {
     // setting limits close to actual heap sizes.
     allocated_size_limit_for_check_ = 0;
 
-    RecordEmbedderSpeed(isolate_->heap()->tracer(),
-                        stats_collector_->marking_time(), used_size_);
+    RecordEmbedderMarkingSpeed(isolate_->heap()->tracer(),
+                               stats_collector_->marking_time(), used_size_);
   }
 }
 
@@ -953,12 +1021,7 @@ void CppHeap::CompactAndSweep() {
             ? cppgc::internal::SweepingConfig::FreeMemoryHandling::
                   kDiscardWherePossible
             : cppgc::internal::SweepingConfig::FreeMemoryHandling::
-                  kDoNotDiscard,
-        // CppHeap is initialized before V8 flags are necessarily set which
-        // prohibits us from reading the flag at creation.
-        v8_flags.cppheap_optimize_sweep_for_mutator
-            ? cppgc::internal::SweepingStrategy::kMinimizeMutatorInterference
-            : cppgc::internal::SweepingStrategy::kMinimizeMemory};
+                  kDoNotDiscard};
     DCHECK_IMPLIES(!isolate_,
                    SweepingType::kAtomic == sweeping_config.sweeping_type);
     sweeper().Start(sweeping_config);
@@ -1044,14 +1107,15 @@ void CppHeap::CollectGarbageForTesting(CollectionType collection_type,
     // Perform an atomic GC, with starting incremental/concurrent marking and
     // immediately finalizing the garbage collection.
     if (!IsMarking()) {
-      InitializeMarking(collection_type, GarbageCollectionFlagValues::kForced);
+      InitializeMarking(collection_type, nullptr,
+                        GarbageCollectionFlagValues::kForced);
       StartMarking();
     }
     EnterFinalPause(stack_state);
     EnterProcessGlobalAtomicPause();
-    CHECK(AdvanceTracing(v8::base::TimeDelta::Max()));
+    CHECK(AdvanceMarking(v8::base::TimeDelta::Max(), SIZE_MAX));
     if (FinishConcurrentMarkingIfNeeded()) {
-      CHECK(AdvanceTracing(v8::base::TimeDelta::Max()));
+      CHECK(AdvanceMarking(v8::base::TimeDelta::Max(), SIZE_MAX));
     }
     FinishMarkingAndProcessWeakness();
     CompactAndSweep();
@@ -1073,7 +1137,7 @@ void CppHeap::StartIncrementalGarbageCollectionForTesting() {
   DCHECK_NULL(isolate_);
   if (IsMarking()) return;
   force_incremental_marking_for_testing_ = true;
-  InitializeMarking(CollectionType::kMajor,
+  InitializeMarking(CollectionType::kMajor, nullptr,
                     GarbageCollectionFlagValues::kForced);
   StartMarking();
   force_incremental_marking_for_testing_ = false;
@@ -1225,6 +1289,9 @@ void CppHeap::CollectGarbage(cppgc::internal::GCConfig config) {
           : GCFlag::kNoFlags;
   isolate_->heap()->CollectAllGarbage(
       flags, GarbageCollectionReason::kCppHeapAllocationFailure);
+  DCHECK_IMPLIES(
+      config.sweeping_type == cppgc::internal::GCConfig::SweepingType::kAtomic,
+      !sweeper_.IsSweepingInProgress());
 }
 
 std::optional<cppgc::EmbedderStackState> CppHeap::overridden_stack_state()
@@ -1302,5 +1369,15 @@ bool CppHeap::IsGCForbidden() const {
          HeapBase::IsGCForbidden();
 }
 
+bool CppHeap::CurrentThreadIsHeapThread() const {
+  if (!is_detached_ && isolate_ &&
+      V8_UNLIKELY(isolate_->was_locker_ever_used())) {
+    // If v8::Locker has been used, we only check if the isolate is now locked
+    // by the current thread.
+    return isolate_->thread_manager()->IsLockedByCurrentThread();
+  }
+  return HeapBase::CurrentThreadIsHeapThread();
+}
+
 }  // namespace internal
 }  // namespace v8
