@@ -40,7 +40,7 @@ class IC {
   // Compute the current IC state based on the target stub, lookup_start_object
   // and name.
   void UpdateState(DirectHandle<Object> lookup_start_object,
-                   Handle<Object> name);
+                   DirectHandle<Object> name);
 
   bool RecomputeHandlerForName(DirectHandle<Object> name);
   void MarkRecomputeHandler(DirectHandle<Object> name) {
@@ -64,7 +64,7 @@ class IC {
 
   static inline bool IsHandler(Tagged<MaybeObject> object);
 
-  // Nofity the IC system that a feedback has changed.
+  // Notify the IC system that a feedback has changed.
   static void OnFeedbackChanged(Isolate* isolate, Tagged<FeedbackVector> vector,
                                 FeedbackSlot slot, const char* reason);
 
@@ -83,39 +83,41 @@ class IC {
   // Configure for most states.
   bool ConfigureVectorState(IC::State new_state, DirectHandle<Object> key);
   // Configure the vector for MONOMORPHIC.
-  void ConfigureVectorState(Handle<Name> name, DirectHandle<Map> map,
-                            Handle<Object> handler);
-  void ConfigureVectorState(Handle<Name> name, DirectHandle<Map> map,
-                            const MaybeObjectHandle& handler);
+  void ConfigureVectorState(DirectHandle<Name> name, DirectHandle<Map> map,
+                            DirectHandle<Object> handler);
+  void ConfigureVectorState(DirectHandle<Name> name, DirectHandle<Map> map,
+                            const MaybeObjectDirectHandle& handler);
   // Configure the vector for POLYMORPHIC.
-  void ConfigureVectorState(Handle<Name> name, MapHandlesSpan maps,
+  void ConfigureVectorState(DirectHandle<Name> name, MapHandlesSpan maps,
                             MaybeObjectHandles* handlers);
-  void ConfigureVectorState(
-      Handle<Name> name, std::vector<MapAndHandler> const& maps_and_handlers);
+  void ConfigureVectorState(DirectHandle<Name> name,
+                            MapsAndHandlers const& maps_and_handlers);
 
   char TransitionMarkFromState(IC::State state);
   void TraceIC(const char* type, DirectHandle<Object> name);
   void TraceIC(const char* type, DirectHandle<Object> name, State old_state,
                State new_state);
 
-  MaybeHandle<Object> TypeError(MessageTemplate, Handle<Object> object,
-                                Handle<Object> key);
-  MaybeHandle<Object> ReferenceError(Handle<Name> name);
+  MaybeDirectHandle<Object> TypeError(MessageTemplate, Handle<Object> object,
+                                      Handle<Object> key);
+  MaybeDirectHandle<Object> ReferenceError(Handle<Name> name);
 
-  void UpdateMonomorphicIC(const MaybeObjectHandle& handler, Handle<Name> name);
-  bool UpdateMegaDOMIC(const MaybeObjectHandle& handler,
+  void UpdateMonomorphicIC(const MaybeObjectDirectHandle& handler,
+                           DirectHandle<Name> name);
+  bool UpdateMegaDOMIC(const MaybeObjectDirectHandle& handler,
                        DirectHandle<Name> name);
-  bool UpdatePolymorphicIC(Handle<Name> name, const MaybeObjectHandle& handler);
+  bool UpdatePolymorphicIC(DirectHandle<Name> name,
+                           const MaybeObjectDirectHandle& handler);
   void UpdateMegamorphicCache(DirectHandle<Map> map, DirectHandle<Name> name,
-                              const MaybeObjectHandle& handler);
+                              const MaybeObjectDirectHandle& handler);
 
   StubCache* stub_cache();
 
   void CopyICToMegamorphicCache(DirectHandle<Name> name);
   bool IsTransitionOfMonomorphicTarget(Tagged<Map> source_map,
                                        Tagged<Map> target_map);
-  void SetCache(Handle<Name> name, Handle<Object> handler);
-  void SetCache(Handle<Name> name, const MaybeObjectHandle& handler);
+  void SetCache(DirectHandle<Name> name, Handle<Object> handler);
+  void SetCache(DirectHandle<Name> name, const MaybeObjectHandle& handler);
   FeedbackSlotKind kind() const { return kind_; }
   bool IsGlobalIC() const { return IsLoadGlobalIC() || IsStoreGlobalIC(); }
   bool IsLoadIC() const { return IsLoadICKind(kind_); }
@@ -141,7 +143,7 @@ class IC {
 
   void TargetMaps(MapHandles* list) {
     FindTargetMaps();
-    for (Handle<Map> map : target_maps_) {
+    for (DirectHandle<Map> map : target_maps_) {
       list->push_back(map);
     }
   }
@@ -196,9 +198,9 @@ class LoadIC : public IC {
   }
 
   // If receiver is empty, use object as the receiver.
-  V8_WARN_UNUSED_RESULT MaybeHandle<Object> Load(
-      Handle<Object> object, Handle<Name> name, bool update_feedback = true,
-      Handle<Object> receiver = Handle<Object>());
+  V8_WARN_UNUSED_RESULT MaybeDirectHandle<Object> Load(
+      Handle<JSAny> object, Handle<Name> name, bool update_feedback = true,
+      DirectHandle<JSAny> receiver = DirectHandle<JSAny>());
 
  protected:
   // Update the inline cache and the global stub cache based on the
@@ -218,8 +220,8 @@ class LoadGlobalIC : public LoadIC {
                FeedbackSlot slot, FeedbackSlotKind kind)
       : LoadIC(isolate, vector, slot, kind) {}
 
-  V8_WARN_UNUSED_RESULT MaybeHandle<Object> Load(Handle<Name> name,
-                                                 bool update_feedback = true);
+  V8_WARN_UNUSED_RESULT MaybeDirectHandle<Object> Load(
+      Handle<Name> name, bool update_feedback = true);
 };
 
 class KeyedLoadIC : public LoadIC {
@@ -228,19 +230,19 @@ class KeyedLoadIC : public LoadIC {
               FeedbackSlot slot, FeedbackSlotKind kind)
       : LoadIC(isolate, vector, slot, kind) {}
 
-  V8_WARN_UNUSED_RESULT MaybeHandle<Object> Load(Handle<Object> object,
-                                                 Handle<Object> key);
+  V8_WARN_UNUSED_RESULT MaybeDirectHandle<Object> Load(Handle<JSAny> object,
+                                                       Handle<Object> key);
 
  protected:
-  V8_WARN_UNUSED_RESULT MaybeHandle<Object> RuntimeLoad(
-      Handle<Object> object, Handle<Object> key, bool* is_found = nullptr);
+  V8_WARN_UNUSED_RESULT MaybeDirectHandle<Object> RuntimeLoad(
+      DirectHandle<JSAny> object, DirectHandle<Object> key,
+      bool* is_found = nullptr);
 
-  V8_WARN_UNUSED_RESULT MaybeHandle<Object> LoadName(Handle<Object> object,
-                                                     DirectHandle<Object> key,
-                                                     Handle<Name> name);
+  V8_WARN_UNUSED_RESULT MaybeDirectHandle<Object> LoadName(
+      Handle<JSAny> object, DirectHandle<Object> key, Handle<Name> name);
 
   // receiver is HeapObject because it could be a String or a JSObject
-  void UpdateLoadElement(Handle<HeapObject> receiver,
+  void UpdateLoadElement(DirectHandle<HeapObject> receiver,
                          KeyedAccessLoadMode new_load_mode);
 
  private:
@@ -265,8 +267,8 @@ class StoreIC : public IC {
     DCHECK(IsAnyStore());
   }
 
-  V8_WARN_UNUSED_RESULT MaybeHandle<Object> Store(
-      Handle<Object> object, Handle<Name> name, Handle<Object> value,
+  V8_WARN_UNUSED_RESULT MaybeDirectHandle<Object> Store(
+      Handle<JSAny> object, Handle<Name> name, DirectHandle<Object> value,
       StoreOrigin store_origin = StoreOrigin::kNamed);
 
   bool LookupForWrite(LookupIterator* it, DirectHandle<Object> value,
@@ -291,8 +293,8 @@ class StoreGlobalIC : public StoreIC {
                 FeedbackSlot slot, FeedbackSlotKind kind)
       : StoreIC(isolate, vector, slot, kind) {}
 
-  V8_WARN_UNUSED_RESULT MaybeHandle<Object> Store(Handle<Name> name,
-                                                  Handle<Object> value);
+  V8_WARN_UNUSED_RESULT MaybeDirectHandle<Object> Store(
+      Handle<Name> name, DirectHandle<Object> value);
 };
 
 enum KeyedStoreCheckMap { kDontCheckMap, kCheckMap };
@@ -315,9 +317,8 @@ class KeyedStoreIC : public StoreIC {
                FeedbackSlot slot, FeedbackSlotKind kind)
       : StoreIC(isolate, vector, slot, kind) {}
 
-  V8_WARN_UNUSED_RESULT MaybeHandle<Object> Store(Handle<Object> object,
-                                                  Handle<Object> name,
-                                                  Handle<Object> value);
+  V8_WARN_UNUSED_RESULT MaybeDirectHandle<Object> Store(
+      Handle<JSAny> object, Handle<Object> name, DirectHandle<Object> value);
 
  protected:
   void UpdateStoreElement(Handle<Map> receiver_map,
@@ -325,15 +326,16 @@ class KeyedStoreIC : public StoreIC {
                           Handle<Map> new_receiver_map);
 
  private:
-  Handle<Map> ComputeTransitionedMap(Handle<Map> map,
-                                     TransitionMode transition_mode);
+  DirectHandle<Map> ComputeTransitionedMap(Handle<Map> map,
+                                           TransitionMode transition_mode);
 
-  Handle<Object> StoreElementHandler(
-      DirectHandle<Map> receiver_map, KeyedAccessStoreMode store_mode,
-      MaybeHandle<UnionOf<Smi, Cell>> prev_validity_cell = kNullMaybeHandle);
+  Handle<Object> StoreElementHandler(DirectHandle<Map> receiver_map,
+                                     KeyedAccessStoreMode store_mode,
+                                     MaybeDirectHandle<UnionOf<Smi, Cell>>
+                                         prev_validity_cell = kNullMaybeHandle);
 
   void StoreElementPolymorphicHandlers(
-      std::vector<MapAndHandler>* receiver_maps_and_handlers,
+      MapsAndHandlers* receiver_maps_and_handlers,
       KeyedAccessStoreMode store_mode);
 
   friend class IC;
@@ -348,8 +350,9 @@ class StoreInArrayLiteralIC : public KeyedStoreIC {
     DCHECK(IsStoreInArrayLiteralICKind(kind()));
   }
 
-  MaybeHandle<Object> Store(Handle<JSArray> array, Handle<Object> index,
-                            Handle<Object> value);
+  MaybeDirectHandle<Object> Store(DirectHandle<JSArray> array,
+                                  Handle<Object> index,
+                                  DirectHandle<Object> value);
 };
 
 }  // namespace internal
