@@ -44,6 +44,7 @@ const {
   findLongestRegisteredExtension,
   resolveForCJSWithHooks,
   loadSourceForCJSWithHooks,
+  populateCJSExportsFromESM,
 } = require('internal/modules/cjs/loader');
 const { fileURLToPath, pathToFileURL, URL } = require('internal/url');
 let debug = require('internal/util/debuglog').debuglog('esm', (fn) => {
@@ -88,6 +89,7 @@ exports.translators = translators;
 /**
  * Converts a URL to a file path if the URL protocol is 'file:'.
  * @param {string} url - The URL to convert.
+ * @returns {string|URL}
  */
 function errPath(url) {
   const parsed = new URL(url);
@@ -147,9 +149,19 @@ function loadCJSModule(module, source, url, filename, isMain) {
       }
       specifier = `${pathToFileURL(path)}`;
     }
+
+    // FIXME(node:59666) Currently, the ESM loader re-invents require() here for imported CJS and this
+    // requires a separate cache to be populated as well as introducing several quirks. This is not ideal.
     const job = cascadedLoader.getModuleJobForRequireInImportedCJS(specifier, url, importAttributes);
     job.runSync();
-    return cjsCache.get(job.url).exports;
+    const mod = cjsCache.get(job.url);
+    assert(mod, `Imported CJS module ${url} failed to load module ${job.url} using require()`);
+    if (!job.module.synthetic && !mod.loaded) {
+      debug('populateCJSExportsFromESM from require(esm) in imported CJS', url, mod, job.module);
+      populateCJSExportsFromESM(mod, job.module, job.module.getNamespace());
+      mod.loaded = true;
+    }
+    return mod.exports;
   };
   setOwnProperty(requireFn, 'resolve', function resolve(specifier) {
     if (!StringPrototypeStartsWith(specifier, 'node:')) {
@@ -170,13 +182,14 @@ function loadCJSModule(module, source, url, filename, isMain) {
 
 // TODO: can we use a weak map instead?
 const cjsCache = new SafeMap();
+
 /**
  * Creates a ModuleWrap object for a CommonJS module.
  * @param {string} url - The URL of the module.
  * @param {string} source - The source code of the module.
  * @param {boolean} isMain - Whether the module is the main module.
  * @param {string} format - Format of the module.
- * @param {typeof loadCJSModule} [loadCJS=loadCJSModule] - The function to load the CommonJS module.
+ * @param {typeof loadCJSModule} [loadCJS] - The function to load the CommonJS module.
  * @returns {ModuleWrap} The ModuleWrap object for the CommonJS module.
  */
 function createCJSModuleWrap(url, source, isMain, format, loadCJS = loadCJSModule) {
@@ -328,16 +341,17 @@ translators.set('commonjs', function commonjsStrategy(url, source, isMain) {
 /**
  * Get or create an entry in the CJS module cache for the given filename.
  * @param {string} filename CJS module filename
+ * @param {CJSModule} parent The parent CJS module
  * @returns {CJSModule} the cached CJS module entry
  */
-function cjsEmplaceModuleCacheEntry(filename, exportNames) {
+function cjsEmplaceModuleCacheEntry(filename, parent) {
   // TODO: Do we want to keep hitting the user mutable CJS loader here?
   let cjsMod = CJSModule._cache[filename];
   if (cjsMod) {
     return cjsMod;
   }
 
-  cjsMod = new CJSModule(filename);
+  cjsMod = new CJSModule(filename, parent);
   cjsMod.filename = filename;
   cjsMod.paths = CJSModule._nodeModulePaths(cjsMod.path);
   cjsMod[kIsCachedByESMLoader] = true;
@@ -346,11 +360,28 @@ function cjsEmplaceModuleCacheEntry(filename, exportNames) {
   return cjsMod;
 }
 
+/**
+ * Emplace a CJS module cache entry for the given URL.
+ * @param {string} url The module URL
+ * @param {CJSModule} parent The parent CJS module
+ * @returns {CJSModule|undefined} the cached CJS module entry, undefined if url cannot be used to identify a CJS entry.
+ */
+exports.cjsEmplaceModuleCacheEntryForURL = function cjsEmplaceModuleCacheEntryForURL(url, parent) {
+  const filename = urlToFilename(url);
+  if (!filename) {
+    return;
+  }
+  const cjsModule = cjsEmplaceModuleCacheEntry(filename, parent);
+  cjsCache.set(url, cjsModule);
+  return cjsModule;
+};
+
 /**
  * Pre-parses a CommonJS module's exports and re-exports.
  * @param {string} filename - The filename of the module.
  * @param {string} [source] - The source code of the module.
  * @param {string} [format]
+ * @returns {{module: CJSModule, exportNames: string[]}}
  */
 function cjsPreparseModuleExports(filename, source, format) {
   const module = cjsEmplaceModuleCacheEntry(filename);
@@ -492,19 +523,16 @@ translators.set('json', function jsonStrategy(url, source) {
  * >} [[Instance]] slot proxy for WebAssembly Module Record
  */
 const wasmInstances = new SafeWeakMap();
-translators.set('wasm', async function(url, source) {
-  emitExperimentalWarning('Importing WebAssembly modules');
-
+translators.set('wasm', function(url, source) {
   assertBufferSource(source, false, 'load');
 
   debug(`Translating WASMModule ${url}`);
 
   let compiled;
   try {
-    // TODO(joyeecheung): implement a translator that just uses
-    // compiled = new WebAssembly.Module(source) to compile it
-    // synchronously.
-    compiled = await WebAssembly.compile(source);
+    compiled = new WebAssembly.Module(source, {
+      builtins: ['js-string'],
+    });
   } catch (err) {
     err.message = errPath(url) + ': ' + err.message;
     throw err;
@@ -516,6 +544,13 @@ translators.set('wasm', async function(url, source) {
     if (impt.kind === 'global') {
       ArrayPrototypePush(wasmGlobalImports, impt);
     }
+    // Prefix reservations per https://webassembly.github.io/esm-integration/js-api/index.html#parse-a-webassembly-module.
+    if (impt.module.startsWith('wasm-js:')) {
+      throw new WebAssembly.LinkError(`Invalid Wasm import "${impt.module}" in ${url}`);
+    }
+    if (impt.name.startsWith('wasm:') || impt.name.startsWith('wasm-js:')) {
+      throw new WebAssembly.LinkError(`Invalid Wasm import name "${impt.module}" in ${url}`);
+    }
     importsList.add(impt.module);
   }
 
@@ -525,12 +560,16 @@ translators.set('wasm', async function(url, source) {
     if (expt.kind === 'global') {
       wasmGlobalExports.add(expt.name);
     }
+    if (expt.name.startsWith('wasm:') || expt.name.startsWith('wasm-js:')) {
+      throw new WebAssembly.LinkError(`Invalid Wasm export name "${expt.name}" in ${url}`);
+    }
     exportsList.add(expt.name);
   }
 
   const createDynamicModule = require('internal/modules/esm/create_dynamic_module');
 
   const { module } = createDynamicModule([...importsList], [...exportsList], url, (reflect) => {
+    emitExperimentalWarning('Importing WebAssembly module instances');
     for (const impt of importsList) {
       const importNs = reflect.imports[impt];
       const wasmInstance = wasmInstances.get(importNs);
