@@ -3,14 +3,12 @@
 const {
   ArrayFrom,
   ArrayPrototypeFilter,
-  ArrayPrototypeIndexOf,
   ArrayPrototypeJoin,
   ArrayPrototypeMap,
   ArrayPrototypePop,
   ArrayPrototypePush,
   ArrayPrototypeReverse,
   ArrayPrototypeShift,
-  ArrayPrototypeSplice,
   ArrayPrototypeUnshift,
   DateNow,
   FunctionPrototypeCall,
@@ -19,33 +17,40 @@ const {
   MathMax,
   MathMaxApply,
   NumberIsFinite,
+  ObjectDefineProperty,
   ObjectSetPrototypeOf,
   RegExpPrototypeExec,
   SafeStringIterator,
   StringPrototypeCodePointAt,
   StringPrototypeEndsWith,
+  StringPrototypeIncludes,
   StringPrototypeRepeat,
+  StringPrototypeReplaceAll,
   StringPrototypeSlice,
+  StringPrototypeSplit,
   StringPrototypeStartsWith,
-  StringPrototypeTrim,
   Symbol,
   SymbolAsyncIterator,
   SymbolDispose,
 } = primordials;
 
-const { codes: {
-  ERR_INVALID_ARG_VALUE,
-  ERR_USE_AFTER_CLOSE,
-} } = require('internal/errors');
+const {
+  AbortError,
+  codes: {
+    ERR_INVALID_ARG_VALUE,
+    ERR_USE_AFTER_CLOSE,
+  },
+} = require('internal/errors');
 
 const {
   validateAbortSignal,
-  validateArray,
-  validateNumber,
   validateString,
   validateUint32,
 } = require('internal/validators');
-const { kEmptyObject } = require('internal/util');
+const {
+  assignFunctionName,
+  kEmptyObject,
+} = require('internal/util');
 const {
   inspect,
   getStringWidth,
@@ -68,12 +73,19 @@ const {
 } = require('internal/readline/callbacks');
 
 const { StringDecoder } = require('string_decoder');
+const { ReplHistory } = require('internal/repl/history');
 
-const kHistorySize = 30;
 const kMaxUndoRedoStackSize = 2048;
 const kMincrlfDelay = 100;
-// \r\n, \n, or \r followed by something other than \n
-const lineEnding = /\r?\n|\r(?!\n)/g;
+/**
+ * The end of a line is signaled by either one of the following:
+ *  - \r\n
+ *  - \n
+ *  - \r followed by something other than \n
+ *  - \u2028 (Unicode 'LINE SEPARATOR')
+ *  - \u2029 (Unicode 'PARAGRAPH SEPARATOR')
+ */
+const lineEnding = /\r?\n|\r(?!\n)|\u2028|\u2029/g;
 
 const kLineObjectStream = Symbol('line object stream');
 const kQuestionCancel = Symbol('kQuestionCancel');
@@ -85,6 +97,8 @@ const ESCAPE_CODE_TIMEOUT = 500;
 // Max length of the kill ring
 const kMaxLengthOfKillRing = 32;
 
+const kMultilinePrompt = Symbol('| ');
+
 const kAddHistory = Symbol('_addHistory');
 const kBeforeEdit = Symbol('_beforeEdit');
 const kDecoder = Symbol('_decoder');
@@ -96,7 +110,9 @@ const kDeleteWordLeft = Symbol('_deleteWordLeft');
 const kDeleteWordRight = Symbol('_deleteWordRight');
 const kGetDisplayPos = Symbol('_getDisplayPos');
 const kHistoryNext = Symbol('_historyNext');
+const kMoveDownOrHistoryNext = Symbol('_moveDownOrHistoryNext');
 const kHistoryPrev = Symbol('_historyPrev');
+const kMoveUpOrHistoryPrev = Symbol('_moveUpOrHistoryPrev');
 const kInsertString = Symbol('_insertString');
 const kLine = Symbol('_line');
 const kLine_buffer = Symbol('_line_buffer');
@@ -106,11 +122,14 @@ const kMoveCursor = Symbol('_moveCursor');
 const kNormalWrite = Symbol('_normalWrite');
 const kOldPrompt = Symbol('_oldPrompt');
 const kOnLine = Symbol('_onLine');
+const kSetLine = Symbol('_setLine');
 const kPreviousKey = Symbol('_previousKey');
 const kPrompt = Symbol('_prompt');
 const kPushToKillRing = Symbol('_pushToKillRing');
 const kPushToUndoStack = Symbol('_pushToUndoStack');
 const kQuestionCallback = Symbol('_questionCallback');
+const kLastCommandErrored = Symbol('_lastCommandErrored');
+const kQuestionReject = Symbol('_questionReject');
 const kRedo = Symbol('_redo');
 const kRedoStack = Symbol('_redoStack');
 const kRefreshLine = Symbol('_refreshLine');
@@ -122,12 +141,21 @@ const kTabCompleter = Symbol('_tabCompleter');
 const kTtyWrite = Symbol('_ttyWrite');
 const kUndo = Symbol('_undo');
 const kUndoStack = Symbol('_undoStack');
+const kIsMultiline = Symbol('_isMultiline');
 const kWordLeft = Symbol('_wordLeft');
 const kWordRight = Symbol('_wordRight');
 const kWriteToOutput = Symbol('_writeToOutput');
 const kYank = Symbol('_yank');
 const kYanking = Symbol('_yanking');
 const kYankPop = Symbol('_yankPop');
+const kSavePreviousState = Symbol('_savePreviousState');
+const kRestorePreviousState = Symbol('_restorePreviousState');
+const kPreviousLine = Symbol('_previousLine');
+const kPreviousCursor = Symbol('_previousCursor');
+const kPreviousCursorCols = Symbol('_previousCursorCols');
+const kMultilineMove = Symbol('_multilineMove');
+const kPreviousPrevRows = Symbol('_previousPrevRows');
+const kAddNewLineOnTTY = Symbol('_addNewLineOnTTY');
 
 function InterfaceConstructor(input, output, completer, terminal) {
   this[kSawReturnAt] = 0;
@@ -141,9 +169,6 @@ function InterfaceConstructor(input, output, completer, terminal) {
 
   FunctionPrototypeCall(EventEmitter, this);
 
-  let history;
-  let historySize;
-  let removeHistoryDuplicates = false;
   let crlfDelay;
   let prompt = '> ';
   let signal;
@@ -153,14 +178,17 @@ function InterfaceConstructor(input, output, completer, terminal) {
     output = input.output;
     completer = input.completer;
     terminal = input.terminal;
-    history = input.history;
-    historySize = input.historySize;
     signal = input.signal;
+
+    // It is possible to configure the history through the input object
+    const historySize = input.historySize;
+    const history = input.history;
+    const removeHistoryDuplicates = input.removeHistoryDuplicates;
+
     if (input.tabSize !== undefined) {
       validateUint32(input.tabSize, 'tabSize', true);
       this.tabSize = input.tabSize;
     }
-    removeHistoryDuplicates = input.removeHistoryDuplicates;
     if (input.prompt !== undefined) {
       prompt = input.prompt;
     }
@@ -181,24 +209,18 @@ function InterfaceConstructor(input, output, completer, terminal) {
 
     crlfDelay = input.crlfDelay;
     input = input.input;
+
+    input.size = historySize;
+    input.history = history;
+    input.removeHistoryDuplicates = removeHistoryDuplicates;
   }
 
+  this.setupHistoryManager(input);
+
   if (completer !== undefined && typeof completer !== 'function') {
     throw new ERR_INVALID_ARG_VALUE('completer', completer);
   }
 
-  if (history === undefined) {
-    history = [];
-  } else {
-    validateArray(history, 'history');
-  }
-
-  if (historySize === undefined) {
-    historySize = kHistorySize;
-  }
-
-  validateNumber(historySize, 'historySize', 0);
-
   // Backwards compat; check the isTTY prop of the output stream
   //  when `terminal` was not specified
   if (terminal === undefined && !(output === null || output === undefined)) {
@@ -208,13 +230,13 @@ function InterfaceConstructor(input, output, completer, terminal) {
   const self = this;
 
   this.line = '';
+  this[kIsMultiline] = false;
   this[kSubstringSearch] = null;
   this.output = output;
   this.input = input;
   this[kUndoStack] = [];
   this[kRedoStack] = [];
-  this.history = history;
-  this.historySize = historySize;
+  this[kPreviousCursorCols] = -1;
 
   // The kill ring is a global list of blocks of text that were previously
   // killed (deleted). If its size exceeds kMaxLengthOfKillRing, the oldest
@@ -224,7 +246,6 @@ function InterfaceConstructor(input, output, completer, terminal) {
   this[kKillRing] = [];
   this[kKillRingCursor] = 0;
 
-  this.removeHistoryDuplicates = !!removeHistoryDuplicates;
   this.crlfDelay = crlfDelay ?
     MathMax(kMincrlfDelay, crlfDelay) :
     kMincrlfDelay;
@@ -234,7 +255,6 @@ function InterfaceConstructor(input, output, completer, terminal) {
 
   this.terminal = !!terminal;
 
-
   function onerror(err) {
     self.emit('error', err);
   }
@@ -313,8 +333,6 @@ function InterfaceConstructor(input, output, completer, terminal) {
     // Cursor position on the line.
     this.cursor = 0;
 
-    this.historyIndex = -1;
-
     if (output !== null && output !== undefined)
       output.on('resize', onresize);
 
@@ -332,7 +350,7 @@ function InterfaceConstructor(input, output, completer, terminal) {
   }
 
   // Current line
-  this.line = '';
+  this[kSetLine]('');
 
   input.resume();
 }
@@ -341,10 +359,6 @@ ObjectSetPrototypeOf(InterfaceConstructor.prototype, EventEmitter.prototype);
 ObjectSetPrototypeOf(InterfaceConstructor, EventEmitter);
 
 class Interface extends InterfaceConstructor {
-  // eslint-disable-next-line no-useless-constructor
-  constructor(input, output, completer, terminal) {
-    super(input, output, completer, terminal);
-  }
   get columns() {
     if (this.output?.columns) return this.output.columns;
     return Infinity;
@@ -367,6 +381,36 @@ class Interface extends InterfaceConstructor {
     return this[kPrompt];
   }
 
+  setupHistoryManager(options) {
+    this.historyManager = new ReplHistory(this, options);
+
+    if (options.onHistoryFileLoaded) {
+      this.historyManager.initialize(options.onHistoryFileLoaded);
+    }
+
+    ObjectDefineProperty(this, 'history', {
+      __proto__: null, configurable: true, enumerable: true,
+      get() { return this.historyManager.history; },
+      set(newHistory) { return this.historyManager.history = newHistory; },
+    });
+
+    ObjectDefineProperty(this, 'historyIndex', {
+      __proto__: null, configurable: true, enumerable: true,
+      get() { return this.historyManager.index; },
+      set(historyIndex) { return this.historyManager.index = historyIndex; },
+    });
+
+    ObjectDefineProperty(this, 'historySize', {
+      __proto__: null, configurable: true, enumerable: true,
+      get() { return this.historyManager.size; },
+    });
+
+    ObjectDefineProperty(this, 'isFlushing', {
+      __proto__: null, configurable: true, enumerable: true,
+      get() { return this.historyManager.isFlushing; },
+    });
+  }
+
   [kSetRawMode](mode) {
     const wasInRawMode = this.input.isRaw;
 
@@ -406,6 +450,11 @@ class Interface extends InterfaceConstructor {
     }
   }
 
+  [kSetLine](line = '') {
+    this.line = line;
+    this[kIsMultiline] = StringPrototypeIncludes(line, '\n');
+  }
+
   [kOnLine](line) {
     if (this[kQuestionCallback]) {
       const cb = this[kQuestionCallback];
@@ -438,39 +487,7 @@ class Interface extends InterfaceConstructor {
   }
 
   [kAddHistory]() {
-    if (this.line.length === 0) return '';
-
-    // If the history is disabled then return the line
-    if (this.historySize === 0) return this.line;
-
-    // If the trimmed line is empty then return the line
-    if (StringPrototypeTrim(this.line).length === 0) return this.line;
-
-    if (this.history.length === 0 || this.history[0] !== this.line) {
-      if (this.removeHistoryDuplicates) {
-        // Remove older history line if identical to new one
-        const dupIndex = ArrayPrototypeIndexOf(this.history, this.line);
-        if (dupIndex !== -1) ArrayPrototypeSplice(this.history, dupIndex, 1);
-      }
-
-      ArrayPrototypeUnshift(this.history, this.line);
-
-      // Only store so many
-      if (this.history.length > this.historySize)
-        ArrayPrototypePop(this.history);
-    }
-
-    this.historyIndex = -1;
-
-    // The listener could change the history object, possibly
-    // to remove the last added entry if it is sensitive and should
-    // not be persisted in the history, like a password
-    const line = this.history[0];
-
-    // Emit history event to notify listeners of update
-    this.emit('history', this.history);
-
-    return line;
+    return this.historyManager.addHistory(this[kIsMultiline], this[kLastCommandErrored]);
   }
 
   [kRefreshLine]() {
@@ -494,8 +511,19 @@ class Interface extends InterfaceConstructor {
     // erase data
     clearScreenDown(this.output);
 
-    // Write the prompt and the current buffer content.
-    this[kWriteToOutput](line);
+    if (this[kIsMultiline]) {
+      const lines = StringPrototypeSplit(this.line, '\n');
+      // Write first line with normal prompt
+      this[kWriteToOutput](this[kPrompt] + lines[0]);
+
+      // For continuation lines, add the "|" prefix
+      for (let i = 1; i < lines.length; i++) {
+        this[kWriteToOutput](`\n${kMultilinePrompt.description}` + lines[i]);
+      }
+    } else {
+      // Write the prompt and the current buffer content.
+      this[kWriteToOutput](line);
+    }
 
     // Force terminal to allocate a new line
     if (lineCols === 0) {
@@ -532,6 +560,9 @@ class Interface extends InterfaceConstructor {
    * @returns {void | Interface}
    */
   pause() {
+    if (this.closed) {
+      throw new ERR_USE_AFTER_CLOSE('readline');
+    }
     if (this.paused) return;
     this.input.pause();
     this.paused = true;
@@ -544,6 +575,9 @@ class Interface extends InterfaceConstructor {
    * @returns {void | Interface}
    */
   resume() {
+    if (this.closed) {
+      throw new ERR_USE_AFTER_CLOSE('readline');
+    }
     if (!this.paused) return;
     this.input.resume();
     this.paused = false;
@@ -564,6 +598,9 @@ class Interface extends InterfaceConstructor {
    * @returns {void}
    */
   write(d, key) {
+    if (this.closed) {
+      throw new ERR_USE_AFTER_CLOSE('readline');
+    }
     if (this.paused) this.resume();
     if (this.terminal) {
       this[kTtyWrite](d, key);
@@ -628,7 +665,7 @@ class Interface extends InterfaceConstructor {
         this.cursor,
         this.line.length,
       );
-      this.line = beg + c + end;
+      this[kSetLine](beg + c + end);
       this.cursor += c.length;
       this[kRefreshLine]();
     } else {
@@ -676,13 +713,13 @@ class Interface extends InterfaceConstructor {
       this[kInsertString](StringPrototypeSlice(prefix, completeOn.length));
       return;
     } else if (!StringPrototypeStartsWith(completeOn, prefix)) {
-      this.line = StringPrototypeSlice(this.line,
-                                       0,
-                                       this.cursor - completeOn.length) +
+      this[kSetLine](StringPrototypeSlice(this.line,
+                                          0,
+                                          this.cursor - completeOn.length) +
                   prefix +
                   StringPrototypeSlice(this.line,
                                        this.cursor,
-                                       this.line.length);
+                                       this.line.length));
       this.cursor = this.cursor - completeOn.length + prefix.length;
       this[kRefreshLine]();
       return;
@@ -821,7 +858,7 @@ class Interface extends InterfaceConstructor {
   [kDeleteLineLeft]() {
     this[kBeforeEdit](this.line, this.cursor);
     const del = StringPrototypeSlice(this.line, 0, this.cursor);
-    this.line = StringPrototypeSlice(this.line, this.cursor);
+    this[kSetLine](StringPrototypeSlice(this.line, this.cursor));
     this.cursor = 0;
     this[kPushToKillRing](del);
     this[kRefreshLine]();
@@ -830,7 +867,7 @@ class Interface extends InterfaceConstructor {
   [kDeleteLineRight]() {
     this[kBeforeEdit](this.line, this.cursor);
     const del = StringPrototypeSlice(this.line, this.cursor);
-    this.line = StringPrototypeSlice(this.line, 0, this.cursor);
+    this[kSetLine](StringPrototypeSlice(this.line, 0, this.cursor));
     this[kPushToKillRing](del);
     this[kRefreshLine]();
   }
@@ -865,21 +902,34 @@ class Interface extends InterfaceConstructor {
             StringPrototypeSlice(this.line, 0, this.cursor - lastYank.length);
       const tail =
             StringPrototypeSlice(this.line, this.cursor);
-      this.line = head + currentYank + tail;
+      this[kSetLine](head + currentYank + tail);
       this.cursor = head.length + currentYank.length;
       this[kRefreshLine]();
     }
   }
 
+  [kSavePreviousState]() {
+    this[kPreviousLine] = this.line;
+    this[kPreviousCursor] = this.cursor;
+    this[kPreviousPrevRows] = this.prevRows;
+  }
+
+  [kRestorePreviousState]() {
+    this[kSetLine](this[kPreviousLine]);
+    this.cursor = this[kPreviousCursor];
+    this.prevRows = this[kPreviousPrevRows];
+  }
+
   clearLine() {
     this[kMoveCursor](+Infinity);
     this[kWriteToOutput]('\r\n');
-    this.line = '';
+    this[kSetLine]('');
     this.cursor = 0;
     this.prevRows = 0;
   }
 
   [kLine]() {
+    this[kSavePreviousState]();
     const line = this[kAddHistory]();
     this[kUndoStack] = [];
     this[kRedoStack] = [];
@@ -887,6 +937,107 @@ class Interface extends InterfaceConstructor {
     this[kOnLine](line);
   }
 
+
+  // TODO(puskin94): edit [kTtyWrite] to make call this function on a new key combination
+  //                 to make it add a new line in the middle of a "complete" multiline.
+  //                 I tried with shift + enter but it is not detected. Find a new one.
+  //                 Make sure to call this[kSavePreviousState](); && this.clearLine();
+  //                 before calling this[kAddNewLineOnTTY] to simulate what [kLine] is doing.
+
+  // When this function is called, the actual cursor is at the very end of the whole string,
+  // No matter where the new line was entered.
+  // This function should only be used when the output is a TTY
+  [kAddNewLineOnTTY]() {
+    // Restore terminal state and store current line
+    this[kRestorePreviousState]();
+    const originalLine = this.line;
+
+    // Split the line at the current cursor position
+    const beforeCursor = StringPrototypeSlice(this.line, 0, this.cursor);
+    let afterCursor = StringPrototypeSlice(this.line, this.cursor, this.line.length);
+
+    // Add the new line where the cursor is at
+    this[kSetLine](`${beforeCursor}\n${afterCursor}`);
+
+    // To account for the new line
+    this.cursor += 1;
+
+    const hasContentAfterCursor = afterCursor.length > 0;
+    const cursorIsNotOnFirstLine = this.prevRows > 0;
+    let needsRewriteFirstLine = false;
+
+    // Handle cursor positioning based on different scenarios
+    if (hasContentAfterCursor) {
+      const splitBeg = StringPrototypeSplit(beforeCursor, '\n');
+      // Determine if we need to rewrite the first line
+      needsRewriteFirstLine = splitBeg.length < 2;
+
+      // If the cursor is not on the first line
+      if (cursorIsNotOnFirstLine) {
+        const splitEnd = StringPrototypeSplit(afterCursor, '\n');
+
+        // If the cursor when I pressed enter was at least on the second line
+        // I need to completely erase the line where the cursor was pressed because it is possible
+        // That it was pressed in the middle of the line, hence I need to write the whole line.
+        // To achieve that, I need to reach the line above the current line coming from the end
+        const dy = splitEnd.length + 1;
+
+        // Calculate how many Xs we need to move on the right to get to the end of the line
+        const dxEndOfLineAbove = (splitBeg[splitBeg.length - 2] || '').length + kMultilinePrompt.description.length;
+        moveCursor(this.output, dxEndOfLineAbove, -dy);
+
+        // This is the line that was split in the middle
+        // Just add it to the rest of the line that will be printed later
+        afterCursor = `${splitBeg[splitBeg.length - 1]}\n${afterCursor}`;
+      } else {
+        // Otherwise, go to the very beginning of the first line and erase everything
+        const dy = StringPrototypeSplit(originalLine, '\n').length;
+        moveCursor(this.output, 0, -dy);
+      }
+
+      // Erase from the cursor to the end of the line
+      clearScreenDown(this.output);
+
+      if (cursorIsNotOnFirstLine) {
+        this[kWriteToOutput]('\n');
+      }
+    }
+
+    if (needsRewriteFirstLine) {
+      this[kWriteToOutput](`${this[kPrompt]}${beforeCursor}\n${kMultilinePrompt.description}`);
+    } else {
+      this[kWriteToOutput](kMultilinePrompt.description);
+    }
+
+    // Write the rest and restore the cursor to where the user left it
+    if (hasContentAfterCursor) {
+      // Save the cursor pos, we need to come back here
+      const oldCursor = this.getCursorPos();
+
+      // Write everything after the cursor which has been deleted by clearScreenDown
+      const formattedEndContent = StringPrototypeReplaceAll(
+        afterCursor,
+        '\n',
+        `\n${kMultilinePrompt.description}`,
+      );
+
+      this[kWriteToOutput](formattedEndContent);
+
+      const newCursor = this[kGetDisplayPos](this.line);
+
+      // Go back to where the cursor was, with relative movement
+      moveCursor(this.output, oldCursor.cols - newCursor.cols, oldCursor.rows - newCursor.rows);
+
+      // Setting how many rows we have on top of the cursor
+      // Necessary for kRefreshLine
+      this.prevRows = oldCursor.rows;
+    } else {
+      // Setting how many rows we have on top of the cursor
+      // Necessary for kRefreshLine
+      this.prevRows = StringPrototypeSplit(this.line, '\n').length - 1;
+    }
+  }
+
   [kPushToUndoStack](text, cursor) {
     if (ArrayPrototypePush(this[kUndoStack], { text, cursor }) >
         kMaxUndoRedoStackSize) {
@@ -903,7 +1054,7 @@ class Interface extends InterfaceConstructor {
     );
 
     const entry = ArrayPrototypePop(this[kUndoStack]);
-    this.line = entry.text;
+    this[kSetLine](entry.text);
     this.cursor = entry.cursor;
 
     this[kRefreshLine]();
@@ -918,12 +1069,59 @@ class Interface extends InterfaceConstructor {
     );
 
     const entry = ArrayPrototypePop(this[kRedoStack]);
-    this.line = entry.text;
+    this[kSetLine](entry.text);
     this.cursor = entry.cursor;
 
     this[kRefreshLine]();
   }
 
+  [kMultilineMove](direction, splitLines, { rows, cols }) {
+    const curr = splitLines[rows];
+    const down = direction === 1;
+    const adj = splitLines[rows + direction];
+    const promptLen = kMultilinePrompt.description.length;
+    let amountToMove;
+    // Clamp distance to end of current + prompt + next/prev line + newline
+    const clamp = down ?
+      curr.length - cols + promptLen + adj.length + 1 :
+      -cols + 1;
+    const shouldClamp = cols > adj.length + 1;
+
+    if (shouldClamp) {
+      if (this[kPreviousCursorCols] === -1) {
+        this[kPreviousCursorCols] = cols;
+      }
+      amountToMove = clamp;
+    } else {
+      if (down) {
+        amountToMove = curr.length + 1;
+      } else {
+        amountToMove = -adj.length - 1;
+      }
+      if (this[kPreviousCursorCols] !== -1) {
+        if (this[kPreviousCursorCols] <= adj.length) {
+          amountToMove += this[kPreviousCursorCols] - cols;
+          this[kPreviousCursorCols] = -1;
+        } else {
+          amountToMove = clamp;
+        }
+      }
+    }
+
+    this[kMoveCursor](amountToMove);
+  }
+
+  [kMoveDownOrHistoryNext]() {
+    const cursorPos = this.getCursorPos();
+    const splitLines = StringPrototypeSplit(this.line, '\n');
+    if (this[kIsMultiline] && cursorPos.rows < splitLines.length - 1) {
+      this[kMultilineMove](1, splitLines, cursorPos);
+      return;
+    }
+    this[kPreviousCursorCols] = -1;
+    this[kHistoryNext]();
+  }
+
   // TODO(BridgeAR): Add underscores to the search part and a red background in
   // case no match is found. This should only be the visual part and not the
   // actual line content!
@@ -932,49 +1130,32 @@ class Interface extends InterfaceConstructor {
   // <ctrl> + N. Only show this after two/three UPs or DOWNs, not on the first
   // one.
   [kHistoryNext]() {
-    if (this.historyIndex >= 0) {
-      this[kBeforeEdit](this.line, this.cursor);
-      const search = this[kSubstringSearch] || '';
-      let index = this.historyIndex - 1;
-      while (
-        index >= 0 &&
-        (!StringPrototypeStartsWith(this.history[index], search) ||
-          this.line === this.history[index])
-      ) {
-        index--;
-      }
-      if (index === -1) {
-        this.line = search;
-      } else {
-        this.line = this.history[index];
-      }
-      this.historyIndex = index;
-      this.cursor = this.line.length; // Set cursor to end of line.
-      this[kRefreshLine]();
+    if (!this.historyManager.canNavigateToNext()) { return; }
+
+    this[kBeforeEdit](this.line, this.cursor);
+    this[kSetLine](this.historyManager.navigateToNext(this[kSubstringSearch]));
+    this.cursor = this.line.length; // Set cursor to end of line.
+    this[kRefreshLine]();
+  }
+
+  [kMoveUpOrHistoryPrev]() {
+    const cursorPos = this.getCursorPos();
+    if (this[kIsMultiline] && cursorPos.rows > 0) {
+      const splitLines = StringPrototypeSplit(this.line, '\n');
+      this[kMultilineMove](-1, splitLines, cursorPos);
+      return;
     }
+    this[kPreviousCursorCols] = -1;
+    this[kHistoryPrev]();
   }
 
   [kHistoryPrev]() {
-    if (this.historyIndex < this.history.length && this.history.length) {
-      this[kBeforeEdit](this.line, this.cursor);
-      const search = this[kSubstringSearch] || '';
-      let index = this.historyIndex + 1;
-      while (
-        index < this.history.length &&
-        (!StringPrototypeStartsWith(this.history[index], search) ||
-          this.line === this.history[index])
-      ) {
-        index++;
-      }
-      if (index === this.history.length) {
-        this.line = search;
-      } else {
-        this.line = this.history[index];
-      }
-      this.historyIndex = index;
-      this.cursor = this.line.length; // Set cursor to end of line.
-      this[kRefreshLine]();
-    }
+    if (!this.historyManager.canNavigateToPrevious()) { return; }
+
+    this[kBeforeEdit](this.line, this.cursor);
+    this[kSetLine](this.historyManager.navigateToPrevious(this[kSubstringSearch]));
+    this.cursor = this.line.length; // Set cursor to end of line.
+    this[kRefreshLine]();
   }
 
   // Returns the last character's display position of the given string
@@ -983,11 +1164,13 @@ class Interface extends InterfaceConstructor {
     const col = this.columns;
     let rows = 0;
     str = stripVTControlCharacters(str);
+
     for (const char of new SafeStringIterator(str)) {
       if (char === '\n') {
         // Rows must be incremented by 1 even if offset = 0 or col = +Infinity.
         rows += MathCeil(offset / col) || 1;
-        offset = 0;
+        // Only add prefix offset for continuation lines in user input (not prompts)
+        offset = this[kIsMultiline] ? kMultilinePrompt.description.length : 0;
         continue;
       }
       // Tabs must be aligned by an offset of the tab size.
@@ -1006,8 +1189,10 @@ class Interface extends InterfaceConstructor {
         offset += 2;
       }
     }
+
     const cols = offset % col;
     rows += (offset - cols) / col;
+
     return { cols, rows };
   }
 
@@ -1020,8 +1205,8 @@ class Interface extends InterfaceConstructor {
    *   }}
    */
   getCursorPos() {
-    const strBeforeCursor =
-      this[kPrompt] + StringPrototypeSlice(this.line, 0, this.cursor);
+    const strBeforeCursor = this[kPrompt] + StringPrototypeSlice(this.line, 0, this.cursor);
+
     return this[kGetDisplayPos](strBeforeCursor);
   }
 
@@ -1057,6 +1242,7 @@ class Interface extends InterfaceConstructor {
     const previousKey = this[kPreviousKey];
     key ||= kEmptyObject;
     this[kPreviousKey] = key;
+    let shouldResetPreviousCursorCols = true;
 
     if (!key.meta || key.name !== 'y') {
       // Reset yanking state unless we are doing yank pop.
@@ -1070,7 +1256,7 @@ class Interface extends InterfaceConstructor {
       !key.meta &&
       !key.shift
     ) {
-      if (this[kSubstringSearch] === null) {
+      if (this[kSubstringSearch] === null && !this[kIsMultiline]) {
         this[kSubstringSearch] = StringPrototypeSlice(
           this.line,
           0,
@@ -1126,6 +1312,7 @@ class Interface extends InterfaceConstructor {
           } else {
             // This readline instance is finished
             this.close();
+            this[kQuestionReject]?.(new AbortError('Aborted with Ctrl+C'));
           }
           break;
 
@@ -1137,6 +1324,7 @@ class Interface extends InterfaceConstructor {
           if (this.cursor === 0 && this.line.length === 0) {
             // This readline instance is finished
             this.close();
+            this[kQuestionReject]?.(new AbortError('Aborted with Ctrl+D'));
           } else if (this.cursor < this.line.length) {
             this[kDeleteRight]();
           }
@@ -1302,11 +1490,13 @@ class Interface extends InterfaceConstructor {
           break;
 
         case 'up':
-          this[kHistoryPrev]();
+          shouldResetPreviousCursorCols = false;
+          this[kMoveUpOrHistoryPrev]();
           break;
 
         case 'down':
-          this[kHistoryNext]();
+          shouldResetPreviousCursorCols = false;
+          this[kMoveDownOrHistoryNext]();
           break;
 
         case 'tab':
@@ -1341,16 +1531,15 @@ class Interface extends InterfaceConstructor {
           }
       }
     }
+    if (shouldResetPreviousCursorCols) {
+      this[kPreviousCursorCols] = -1;
+    }
   }
 
   /**
    * Creates an `AsyncIterator` object that iterates through
    * each line in the input stream as a string.
-   * @typedef {{
-   *   [Symbol.asyncIterator]: () => InterfaceAsyncIterator,
-   *   next: () => Promise<string>
-   * }} InterfaceAsyncIterator
-   * @returns {InterfaceAsyncIterator}
+   * @returns {AsyncIterableIterator<string>}
    */
   [SymbolAsyncIterator]() {
     if (this[kLineObjectStream] === undefined) {
@@ -1365,6 +1554,9 @@ class Interface extends InterfaceConstructor {
     return this[kLineObjectStream];
   }
 }
+Interface.prototype[SymbolDispose] = assignFunctionName(SymbolDispose, function() {
+  this.close();
+});
 
 module.exports = {
   Interface,
@@ -1381,17 +1573,20 @@ module.exports = {
   kHistoryNext,
   kHistoryPrev,
   kInsertString,
+  kIsMultiline,
   kLine,
   kLine_buffer,
   kMoveCursor,
   kNormalWrite,
   kOldPrompt,
   kOnLine,
+  kSetLine,
   kPreviousKey,
   kPrompt,
   kQuestion,
   kQuestionCallback,
   kQuestionCancel,
+  kQuestionReject,
   kRefreshLine,
   kSawKeyPress,
   kSawReturnAt,
@@ -1402,4 +1597,8 @@ module.exports = {
   kWordLeft,
   kWordRight,
   kWriteToOutput,
+  kMultilinePrompt,
+  kRestorePreviousState,
+  kAddNewLineOnTTY,
+  kLastCommandErrored,
 };
