@@ -36,7 +36,6 @@ const {
   fileURLToPath,
   isURL,
   pathToFileURL,
-  toPathIfFileURL,
   URL,
 } = require('internal/url');
 const {
@@ -210,7 +209,7 @@ class MockModuleContext {
       defaultExport,
       hasDefaultExport,
       namedExports,
-      caller: toPathIfFileURL(caller),
+      caller,
     };
 
     sharedState.mockMap.set(baseURL, config);
@@ -428,7 +427,7 @@ class MockTracker {
    * @param {Function} [original] - The original function to be tracked.
    * @param {Function} [implementation] - An optional replacement function for the original one.
    * @param {object} [options] - Additional tracking options.
-   * @param {number} [options.times=Infinity] - The maximum number of times the mock function can be called.
+   * @param {number} [options.times] - The maximum number of times the mock function can be called.
    * @returns {ProxyConstructor} The mock function tracker.
    */
   fn(
@@ -460,9 +459,9 @@ class MockTracker {
    * @param {string} methodName - The name of the method to be tracked.
    * @param {Function} [implementation] - An optional replacement function for the original method.
    * @param {object} [options] - Additional tracking options.
-   * @param {boolean} [options.getter=false] - Indicates whether this is a getter method.
-   * @param {boolean} [options.setter=false] - Indicates whether this is a setter method.
-   * @param {number} [options.times=Infinity] - The maximum number of times the mock method can be called.
+   * @param {boolean} [options.getter] - Indicates whether this is a getter method.
+   * @param {boolean} [options.setter] - Indicates whether this is a setter method.
+   * @param {number} [options.times] - The maximum number of times the mock method can be called.
    * @returns {ProxyConstructor} The mock method tracker.
    */
   method(
@@ -551,9 +550,9 @@ class MockTracker {
    * @param {string} methodName - The name of the getter method to be mocked.
    * @param {Function} [implementation] - An optional replacement function for the targeted method.
    * @param {object} [options] - Additional tracking options.
-   * @param {boolean} [options.getter=true] - Indicates whether this is a getter method.
-   * @param {boolean} [options.setter=false] - Indicates whether this is a setter method.
-   * @param {number} [options.times=Infinity] - The maximum number of times the mock method can be called.
+   * @param {boolean} [options.getter] - Indicates whether this is a getter method.
+   * @param {boolean} [options.setter] - Indicates whether this is a setter method.
+   * @param {number} [options.times] - The maximum number of times the mock method can be called.
    * @returns {ProxyConstructor} The mock method tracker.
    */
   getter(
@@ -588,12 +587,12 @@ class MockTracker {
    * Mocks a setter method of an object.
    * This function is a syntax sugar for MockTracker.method with options.setter set to true.
    * @param {object} object - The target object.
-   * @param {string} methodName  - The setter method to be mocked.
+   * @param {string} methodName - The setter method to be mocked.
    * @param {Function} [implementation] - An optional replacement function for the targeted method.
    * @param {object} [options] - Additional tracking options.
-   * @param {boolean} [options.getter=false] - Indicates whether this is a getter method.
-   * @param {boolean} [options.setter=true] - Indicates whether this is a setter method.
-   * @param {number} [options.times=Infinity] - The maximum number of times the mock method can be called.
+   * @param {boolean} [options.getter] - Indicates whether this is a getter method.
+   * @param {boolean} [options.setter] - Indicates whether this is a setter method.
+   * @param {number} [options.times] - The maximum number of times the mock method can be called.
    * @returns {ProxyConstructor} The mock method tracker.
    */
   setter(
