@@ -38,6 +38,7 @@ const {
   ERR_VM_MODULE_DIFFERENT_CONTEXT,
   ERR_VM_MODULE_CANNOT_CREATE_CACHED_DATA,
   ERR_VM_MODULE_LINK_FAILURE,
+  ERR_MODULE_LINK_MISMATCH,
   ERR_VM_MODULE_NOT_MODULE,
   ERR_VM_MODULE_STATUS,
 } = require('internal/errors').codes;
@@ -50,6 +51,7 @@ const {
   validateUint32,
   validateString,
   validateThisInternalField,
+  validateArray,
 } = require('internal/validators');
 
 const binding = internalBinding('module_wrap');
@@ -369,6 +371,37 @@ class SourceTextModule extends Module {
     }
   }
 
+  linkRequests(modules) {
+    validateThisInternalField(this, kWrap, 'SourceTextModule');
+    if (this.status !== 'unlinked') {
+      throw new ERR_VM_MODULE_STATUS('must be unlinked');
+    }
+    validateArray(modules, 'modules');
+    if (modules.length !== this.#moduleRequests.length) {
+      throw new ERR_MODULE_LINK_MISMATCH(
+        `Expected ${this.#moduleRequests.length} modules, got ${modules.length}`,
+      );
+    }
+    const moduleWraps = ArrayPrototypeMap(modules, (module) => {
+      if (!isModule(module)) {
+        throw new ERR_VM_MODULE_NOT_MODULE();
+      }
+      if (module.context !== this.context) {
+        throw new ERR_VM_MODULE_DIFFERENT_CONTEXT();
+      }
+      return module[kWrap];
+    });
+    this[kWrap].link(moduleWraps);
+  }
+
+  instantiate() {
+    validateThisInternalField(this, kWrap, 'SourceTextModule');
+    if (this.status !== 'unlinked') {
+      throw new ERR_VM_MODULE_STATUS('must be unlinked');
+    }
+    this[kWrap].instantiate();
+  }
+
   get dependencySpecifiers() {
     this.#dependencySpecifiers ??= ObjectFreeze(
       ArrayPrototypeMap(this.#moduleRequests, (request) => request.specifier));
@@ -435,10 +468,15 @@ class SyntheticModule extends Module {
       context,
       identifier,
     });
+    // A synthetic module does not have dependencies.
+    this[kWrap].link([]);
+    this[kWrap].instantiate();
   }
 
-  [kLink]() {
-    /** nothing to do for synthetic modules */
+  link() {
+    validateThisInternalField(this, kWrap, 'SyntheticModule');
+    // No-op for synthetic modules
+    // Do not invoke super.link() as it will throw an error.
   }
 
   setExport(name, value) {
