@@ -25,6 +25,7 @@ const {
   ArrayIsArray,
   ArrayPrototypePop,
   ArrayPrototypePush,
+  ArrayPrototypeReduce,
   Error,
   ErrorCaptureStackTrace,
   FunctionPrototypeBind,
@@ -36,6 +37,8 @@ const {
   ObjectSetPrototypeOf,
   ObjectValues,
   ReflectApply,
+  RegExp,
+  RegExpPrototypeSymbolReplace,
   StringPrototypeToWellFormed,
 } = primordials;
 
@@ -100,25 +103,27 @@ function lazyAbortController() {
 let internalDeepEqual;
 
 /**
- * @param {string} code
+ * @param {string} [code]
  * @returns {string}
  */
 function escapeStyleCode(code) {
+  if (code === undefined) return '';
   return `\u001b[${code}m`;
 }
 
 /**
  * @param {string | string[]} format
  * @param {string} text
- * @param {object} [options={}]
- * @param {boolean} [options.validateStream=true] - Whether to validate the stream.
- * @param {Stream} [options.stream=process.stdout] - The stream used for validation.
+ * @param {object} [options]
+ * @param {boolean} [options.validateStream] - Whether to validate the stream.
+ * @param {Stream} [options.stream] - The stream used for validation.
  * @returns {string}
  */
 function styleText(format, text, { validateStream = true, stream = process.stdout } = {}) {
   validateString(text, 'text');
   validateBoolean(validateStream, 'options.validateStream');
 
+  let skipColorize;
   if (validateStream) {
     if (
       !isReadableStream(stream) &&
@@ -127,40 +132,73 @@ function styleText(format, text, { validateStream = true, stream = process.stdou
     ) {
       throw new ERR_INVALID_ARG_TYPE('stream', ['ReadableStream', 'WritableStream', 'Stream'], stream);
     }
+
+    // If the stream is falsy or should not be colorized, set skipColorize to true
+    skipColorize = !lazyUtilColors().shouldColorize(stream);
   }
 
-  if (ArrayIsArray(format)) {
-    let left = '';
-    let right = '';
-    for (const key of format) {
-      const formatCodes = inspect.colors[key];
-      if (formatCodes == null) {
-        validateOneOf(key, 'format', ObjectKeys(inspect.colors));
-      }
-      left += escapeStyleCode(formatCodes[0]);
-      right = `${escapeStyleCode(formatCodes[1])}${right}`;
+  // If the format is not an array, convert it to an array
+  const formatArray = ArrayIsArray(format) ? format : [format];
+
+  const codes = [];
+  for (const key of formatArray) {
+    if (key === 'none') continue;
+    const formatCodes = inspect.colors[key];
+    // If the format is not a valid style, throw an error
+    if (formatCodes == null) {
+      validateOneOf(key, 'format', ObjectKeys(inspect.colors));
     }
-
-    return `${left}${text}${right}`;
+    if (skipColorize) continue;
+    ArrayPrototypePush(codes, formatCodes);
   }
 
-  const formatCodes = inspect.colors[format];
-  if (formatCodes == null) {
-    validateOneOf(format, 'format', ObjectKeys(inspect.colors));
-  }
-
-  // Check colorize only after validating arg type and value
-  if (
-    validateStream &&
-    (
-      !stream ||
-      !lazyUtilColors().shouldColorize(stream)
-    )
-  ) {
+  if (skipColorize) {
     return text;
   }
 
-  return `${escapeStyleCode(formatCodes[0])}${text}${escapeStyleCode(formatCodes[1])}`;
+  // Build opening codes
+  let openCodes = '';
+  for (let i = 0; i < codes.length; i++) {
+    openCodes += escapeStyleCode(codes[i][0]);
+  }
+
+  // Process the text to handle nested styles
+  let processedText;
+  if (codes.length > 0) {
+    processedText = ArrayPrototypeReduce(
+      codes,
+      (text, code) => RegExpPrototypeSymbolReplace(
+        // Find the reset code
+        new RegExp(`\\u001b\\[${code[1]}m`, 'g'),
+        text,
+        (match, offset) => {
+          // Check if there's more content after this reset
+          if (offset + match.length < text.length) {
+            if (
+              code[0] === inspect.colors.dim[0] ||
+              code[0] === inspect.colors.bold[0]
+            ) {
+              // Dim and bold are not mutually exclusive, so we need to reapply
+              return `${match}${escapeStyleCode(code[0])}`;
+            }
+            return escapeStyleCode(code[0]);
+          }
+          return match;
+        },
+      ),
+      text,
+    );
+  } else {
+    processedText = text;
+  }
+
+  // Build closing codes in reverse order
+  let closeCodes = '';
+  for (let i = codes.length - 1; i >= 0; i--) {
+    closeCodes += escapeStyleCode(codes[i][1]);
+  }
+
+  return `${openCodes}${processedText}${closeCodes}`;
 }
 
 /**
@@ -172,10 +210,10 @@ function styleText(format, text, { validateStream = true, stream = process.stdou
  * functions as prototype setup using normal JavaScript does not work as
  * expected during bootstrapping (see mirror.js in r114903).
  * @param {Function} ctor Constructor function which needs to inherit the
- *     prototype.
+ *   prototype.
  * @param {Function} superCtor Constructor function to inherit prototype from.
  * @throws {TypeError} Will error if either constructor is null, or if
- *     the super constructor lacks a prototype.
+ * the super constructor lacks a prototype.
  */
 function inherits(ctor, superCtor) {
 
@@ -204,7 +242,7 @@ function inherits(ctor, superCtor) {
  * @template S
  * @param {T} target
  * @param {S} source
- * @returns {S extends null ? T : (T & S)}
+ * @returns {(T & S) | null}
  */
 function _extend(target, source) {
   // Don't do anything if source isn't an object
@@ -231,12 +269,9 @@ const callbackifyOnRejected = (reason, cb) => {
 };
 
 /**
- * @template {(...args: any[]) => Promise<any>} T
- * @param {T} original
- * @returns {T extends (...args: infer TArgs) => Promise<infer TReturn> ?
- *   ((...params: [...TArgs, ((err: Error, ret: TReturn) => any)]) => void) :
- *   never
- * }
+ * Converts a Promise-returning function to callback style
+ * @param {Function} original
+ * @returns {Function}
  */
 function callbackify(original) {
   validateFunction(original, 'original');
@@ -337,7 +372,7 @@ const lazySourceMap = getLazy(() => require('internal/source_map/source_map_cach
 /**
  * @typedef {object} CallSite // The call site
  * @property {string} scriptName // The name of the resource that contains the
- * script for the function for this StackFrame
+ *   script for the function for this StackFrame
  * @property {string} functionName // The name of the function associated with this stack frame
  * @property {number} lineNumber // The number, 1-based, of the line for the associate function call
  * @property {number} columnNumber // The 1-based column offset on the line for the associated function call
@@ -348,10 +383,10 @@ const lazySourceMap = getLazy(() => require('internal/source_map/source_map_cach
  * @returns {CallSite | undefined} // The reconstructed call site object
  */
 function reconstructCallSite(callSite) {
-  const { scriptName, lineNumber, column } = callSite;
+  const { scriptName, lineNumber, columnNumber } = callSite;
   const sourceMap = lazySourceMap().findSourceMap(scriptName);
   if (!sourceMap) return;
-  const entry = sourceMap.findEntry(lineNumber - 1, column - 1);
+  const entry = sourceMap.findEntry(lineNumber - 1, columnNumber - 1);
   if (!entry?.originalSource) return;
   return {
     __proto__: null,
@@ -360,6 +395,7 @@ function reconstructCallSite(callSite) {
     scriptName: entry.originalSource,
     lineNumber: entry.originalLine + 1,
     column: entry.originalColumn + 1,
+    columnNumber: entry.originalColumn + 1,
   };
 }
 
@@ -367,7 +403,7 @@ function reconstructCallSite(callSite) {
  *
  * The call site array to map
  * @param {CallSite[]} callSites
- * Array of objects with the reconstructed call site
+ *   Array of objects with the reconstructed call site
  * @returns {CallSite[]}
  */
 function mapCallSite(callSites) {
@@ -398,7 +434,9 @@ function getCallSites(frameCount = 10, options) {
       // If frameCount is an object, it is the options object
       options = frameCount;
       validateObject(options, 'options');
-      validateBoolean(options.sourceMap, 'options.sourceMap');
+      if (options.sourceMap !== undefined) {
+        validateBoolean(options.sourceMap, 'options.sourceMap');
+      }
       frameCount = 10;
     } else {
       // If options is not provided, set it to an empty object
@@ -407,7 +445,9 @@ function getCallSites(frameCount = 10, options) {
   } else {
     // If options is provided, validate it
     validateObject(options, 'options');
-    validateBoolean(options.sourceMap, 'options.sourceMap');
+    if (options.sourceMap !== undefined) {
+      validateBoolean(options.sourceMap, 'options.sourceMap');
+    }
   }
 
   // Using kDefaultMaxCallStackSizeToCapture as reference
@@ -489,3 +529,15 @@ defineLazyProperties(
   'internal/mime',
   ['MIMEType', 'MIMEParams'],
 );
+
+defineLazyProperties(
+  module.exports,
+  'internal/util/diff',
+  ['diff'],
+);
+
+defineLazyProperties(
+  module.exports,
+  'internal/util/trace_sigint',
+  ['setTraceSigInt'],
+);
