@@ -1,12 +1,13 @@
 #include "crypto/crypto_keys.h"
-#include "crypto/crypto_common.h"
-#include "crypto/crypto_dsa.h"
-#include "crypto/crypto_ec.h"
-#include "crypto/crypto_dh.h"
-#include "crypto/crypto_rsa.h"
-#include "crypto/crypto_util.h"
 #include "async_wrap-inl.h"
 #include "base_object-inl.h"
+#include "crypto/crypto_common.h"
+#include "crypto/crypto_dh.h"
+#include "crypto/crypto_dsa.h"
+#include "crypto/crypto_ec.h"
+#include "crypto/crypto_ml_dsa.h"
+#include "crypto/crypto_rsa.h"
+#include "crypto/crypto_util.h"
 #include "env-inl.h"
 #include "memory_tracker-inl.h"
 #include "node.h"
@@ -18,6 +19,12 @@
 
 namespace node {
 
+using ncrypto::BIOPointer;
+using ncrypto::ECKeyPointer;
+using ncrypto::EVPKeyCtxPointer;
+using ncrypto::EVPKeyPointer;
+using ncrypto::MarkPopErrorOnReturn;
+using ncrypto::PKCS8Pointer;
 using v8::Array;
 using v8::Context;
 using v8::Function;
@@ -26,7 +33,6 @@ using v8::FunctionTemplate;
 using v8::Int32;
 using v8::Isolate;
 using v8::Just;
-using v8::JustVoid;
 using v8::Local;
 using v8::Maybe;
 using v8::MaybeLocal;
@@ -41,11 +47,11 @@ using v8::Value;
 
 namespace crypto {
 namespace {
-Maybe<ncrypto::EVPKeyPointer::AsymmetricKeyEncodingConfig>
-GetKeyFormatAndTypeFromJs(const FunctionCallbackInfo<Value>& args,
-                          unsigned int* offset,
-                          KeyEncodingContext context) {
-  ncrypto::EVPKeyPointer::AsymmetricKeyEncodingConfig config;
+Maybe<EVPKeyPointer::AsymmetricKeyEncodingConfig> GetKeyFormatAndTypeFromJs(
+    const FunctionCallbackInfo<Value>& args,
+    unsigned int* offset,
+    KeyEncodingContext context) {
+  EVPKeyPointer::AsymmetricKeyEncodingConfig config;
   // During key pair generation, it is possible not to specify a key encoding,
   // which will lead to a key object being returned.
   if (args[*offset]->IsUndefined()) {
@@ -56,19 +62,19 @@ GetKeyFormatAndTypeFromJs(const FunctionCallbackInfo<Value>& args,
     config.output_key_object = false;
 
     CHECK(args[*offset]->IsInt32());
-    config.format = static_cast<ncrypto::EVPKeyPointer::PKFormatType>(
+    config.format = static_cast<EVPKeyPointer::PKFormatType>(
         args[*offset].As<Int32>()->Value());
 
     if (args[*offset + 1]->IsInt32()) {
-      config.type = static_cast<ncrypto::EVPKeyPointer::PKEncodingType>(
+      config.type = static_cast<EVPKeyPointer::PKEncodingType>(
           args[*offset + 1].As<Int32>()->Value());
     } else {
       CHECK((context == kKeyContextInput &&
-             config.format == ncrypto::EVPKeyPointer::PKFormatType::PEM) ||
+             config.format == EVPKeyPointer::PKFormatType::PEM) ||
             (context == kKeyContextGenerate &&
-             config.format == ncrypto::EVPKeyPointer::PKFormatType::JWK));
+             config.format == EVPKeyPointer::PKFormatType::JWK));
       CHECK(args[*offset + 1]->IsNullOrUndefined());
-      config.type = ncrypto::EVPKeyPointer::PKEncodingType::PKCS1;
+      config.type = EVPKeyPointer::PKEncodingType::PKCS1;
     }
   }
 
@@ -76,19 +82,20 @@ GetKeyFormatAndTypeFromJs(const FunctionCallbackInfo<Value>& args,
   return Just(config);
 }
 
-MaybeLocal<Value> BIOToStringOrBuffer(
+MaybeLocal<Value> ToV8Value(
     Environment* env,
     const BIOPointer& bio,
-    const ncrypto::EVPKeyPointer::AsymmetricKeyEncodingConfig& config) {
+    const EVPKeyPointer::AsymmetricKeyEncodingConfig& config) {
+  if (!bio) return {};
   BUF_MEM* bptr = bio;
-  if (config.format == ncrypto::EVPKeyPointer::PKFormatType::PEM) {
+  if (config.format == EVPKeyPointer::PKFormatType::PEM) {
     // PEM is an ASCII format, so we will return it as a string.
     return String::NewFromUtf8(
                env->isolate(), bptr->data, NewStringType::kNormal, bptr->length)
         .FromMaybe(Local<Value>());
   }
 
-  CHECK_EQ(config.format, ncrypto::EVPKeyPointer::PKFormatType::DER);
+  CHECK_EQ(config.format, EVPKeyPointer::PKFormatType::DER);
   // DER is binary, return it as a buffer.
   return Buffer::Copy(env, bptr->data, bptr->length).FromMaybe(Local<Value>());
 }
@@ -96,12 +103,10 @@ MaybeLocal<Value> BIOToStringOrBuffer(
 MaybeLocal<Value> WritePrivateKey(
     Environment* env,
     const EVPKeyPointer& pkey,
-    const ncrypto::EVPKeyPointer::PrivateKeyEncodingConfig& config) {
-  CHECK(pkey);
+    const EVPKeyPointer::PrivateKeyEncodingConfig& config) {
+  if (!pkey) return {};
   auto res = pkey.writePrivateKey(config);
-  if (res) {
-    return BIOToStringOrBuffer(env, std::move(res.value), config);
-  }
+  if (res) return ToV8Value(env, std::move(res.value), config);
 
   ThrowCryptoError(
       env, res.openssl_error.value_or(0), "Failed to encode private key");
@@ -111,48 +116,31 @@ MaybeLocal<Value> WritePrivateKey(
 MaybeLocal<Value> WritePublicKey(
     Environment* env,
     const EVPKeyPointer& pkey,
-    const ncrypto::EVPKeyPointer::PublicKeyEncodingConfig& config) {
-  CHECK(pkey);
+    const EVPKeyPointer::PublicKeyEncodingConfig& config) {
+  if (!pkey) return {};
   auto res = pkey.writePublicKey(config);
-  if (res) {
-    return BIOToStringOrBuffer(env, res.value, config);
-  }
+  if (res) return ToV8Value(env, res.value, config);
 
   ThrowCryptoError(
       env, res.openssl_error.value_or(0), "Failed to encode public key");
   return MaybeLocal<Value>();
 }
 
-Maybe<void> ExportJWKSecretKey(Environment* env,
-                               const KeyObjectData& key,
-                               Local<Object> target) {
+bool ExportJWKSecretKey(Environment* env,
+                        const KeyObjectData& key,
+                        Local<Object> target) {
   CHECK_EQ(key.GetKeyType(), kKeyTypeSecret);
 
-  Local<Value> error;
   Local<Value> raw;
-  MaybeLocal<Value> key_data = StringBytes::Encode(env->isolate(),
-                                                   key.GetSymmetricKey(),
-                                                   key.GetSymmetricKeySize(),
-                                                   BASE64URL,
-                                                   &error);
-  if (!key_data.ToLocal(&raw)) {
-    CHECK(!error.IsEmpty());
-    env->isolate()->ThrowException(error);
-    return Nothing<void>();
-  }
-
-  if (target->Set(
-          env->context(),
-          env->jwk_kty_string(),
-          env->jwk_oct_string()).IsNothing() ||
-      target->Set(
-          env->context(),
-          env->jwk_k_string(),
-          raw).IsNothing()) {
-    return Nothing<void>();
-  }
-
-  return JustVoid();
+  return StringBytes::Encode(env->isolate(),
+                             key.GetSymmetricKey(),
+                             key.GetSymmetricKeySize(),
+                             BASE64URL)
+             .ToLocal(&raw) &&
+         target
+             ->Set(env->context(), env->jwk_kty_string(), env->jwk_oct_string())
+             .IsJust() &&
+         target->Set(env->context(), env->jwk_k_string(), raw).IsJust();
 }
 
 KeyObjectData ImportJWKSecretKey(Environment* env, Local<Object> jwk) {
@@ -168,10 +156,10 @@ KeyObjectData ImportJWKSecretKey(Environment* env, Local<Object> jwk) {
       ByteSource::FromEncodedString(env, key.As<String>()));
 }
 
-Maybe<void> ExportJWKAsymmetricKey(Environment* env,
-                                   const KeyObjectData& key,
-                                   Local<Object> target,
-                                   bool handleRsaPss) {
+bool ExportJWKAsymmetricKey(Environment* env,
+                            const KeyObjectData& key,
+                            Local<Object> target,
+                            bool handleRsaPss) {
   switch (key.GetAsymmetricKey().id()) {
     case EVP_PKEY_RSA_PSS: {
       if (handleRsaPss) return ExportJWKRsaKey(env, key, target);
@@ -189,9 +177,17 @@ Maybe<void> ExportJWKAsymmetricKey(Environment* env,
       // Fall through
     case EVP_PKEY_X448:
       return ExportJWKEdKey(env, key, target);
+#if OPENSSL_WITH_PQC
+    case EVP_PKEY_ML_DSA_44:
+      // Fall through
+    case EVP_PKEY_ML_DSA_65:
+      // Fall through
+    case EVP_PKEY_ML_DSA_87:
+      return ExportJwkMlDsaKey(env, key, target);
+#endif
   }
   THROW_ERR_CRYPTO_JWK_UNSUPPORTED_KEY_TYPE(env);
-  return Nothing<void>();
+  return false;
 }
 
 KeyObjectData ImportJWKAsymmetricKey(Environment* env,
@@ -210,40 +206,44 @@ KeyObjectData ImportJWKAsymmetricKey(Environment* env,
   return {};
 }
 
-Maybe<void> GetSecretKeyDetail(Environment* env,
-                               const KeyObjectData& key,
-                               Local<Object> target) {
+bool GetSecretKeyDetail(Environment* env,
+                        const KeyObjectData& key,
+                        Local<Object> target) {
   // For the secret key detail, all we care about is the length,
   // converted to bits.
-  size_t length = key.GetSymmetricKeySize() * CHAR_BIT;
-  if (target
-          ->Set(env->context(),
-                env->length_string(),
-                Number::New(env->isolate(), static_cast<double>(length)))
-          .IsNothing()) {
-    return Nothing<void>();
-  }
-  return JustVoid();
+  return target
+      ->Set(env->context(),
+            env->length_string(),
+            Number::New(
+                env->isolate(),
+                static_cast<double>(key.GetSymmetricKeySize() * CHAR_BIT)))
+      .IsJust();
 }
 
-Maybe<void> GetAsymmetricKeyDetail(Environment* env,
-                                   const KeyObjectData& key,
-                                   Local<Object> target) {
+bool GetAsymmetricKeyDetail(Environment* env,
+                            const KeyObjectData& key,
+                            Local<Object> target) {
+  if (!key) {
+    THROW_ERR_CRYPTO_OPERATION_FAILED(env);
+    return false;
+  }
   switch (key.GetAsymmetricKey().id()) {
     case EVP_PKEY_RSA:
       // Fall through
+    case EVP_PKEY_RSA2:
+      // Fall through
     case EVP_PKEY_RSA_PSS: return GetRsaKeyDetail(env, key, target);
     case EVP_PKEY_DSA: return GetDsaKeyDetail(env, key, target);
     case EVP_PKEY_EC: return GetEcKeyDetail(env, key, target);
     case EVP_PKEY_DH: return GetDhKeyDetail(env, key, target);
   }
   THROW_ERR_CRYPTO_INVALID_KEYTYPE(env);
-  return Nothing<void>();
+  return false;
 }
 
 KeyObjectData TryParsePrivateKey(
     Environment* env,
-    const ncrypto::EVPKeyPointer::PrivateKeyEncodingConfig& config,
+    const EVPKeyPointer::PrivateKeyEncodingConfig& config,
     const ncrypto::Buffer<const unsigned char>& buffer) {
   auto res = EVPKeyPointer::TryParsePrivateKey(config, buffer);
   if (res) {
@@ -260,79 +260,94 @@ KeyObjectData TryParsePrivateKey(
   return {};
 }
 
-// This maps true to JustVoid and false to Nothing<void>().
-static inline Maybe<void> NothingIfFalse(bool b) {
-  return b ? JustVoid() : Nothing<void>();
+bool ExportJWKInner(Environment* env,
+                    const KeyObjectData& key,
+                    Local<Value> result,
+                    bool handleRsaPss) {
+  return key.GetKeyType() == kKeyTypeSecret
+             ? ExportJWKSecretKey(env, key, result.As<Object>())
+             : ExportJWKAsymmetricKey(
+                   env, key, result.As<Object>(), handleRsaPss);
 }
 
-Maybe<void> ExportJWKInner(Environment* env,
-                           const KeyObjectData& key,
-                           Local<Value> result,
-                           bool handleRsaPss) {
-  switch (key.GetKeyType()) {
-    case kKeyTypeSecret:
-      return ExportJWKSecretKey(env, key, result.As<Object>());
-    case kKeyTypePublic:
-      // Fall through
-    case kKeyTypePrivate:
-      return ExportJWKAsymmetricKey(
-        env, key, result.As<Object>(), handleRsaPss);
-    default:
-      UNREACHABLE();
+int GetNidFromName(const char* name) {
+  int nid;
+  if (strcmp(name, "Ed25519") == 0) {
+    nid = EVP_PKEY_ED25519;
+  } else if (strcmp(name, "Ed448") == 0) {
+    nid = EVP_PKEY_ED448;
+  } else if (strcmp(name, "X25519") == 0) {
+    nid = EVP_PKEY_X25519;
+  } else if (strcmp(name, "X448") == 0) {
+    nid = EVP_PKEY_X448;
+#if OPENSSL_WITH_PQC
+  } else if (strcmp(name, "ML-DSA-44") == 0) {
+    nid = EVP_PKEY_ML_DSA_44;
+  } else if (strcmp(name, "ML-DSA-65") == 0) {
+    nid = EVP_PKEY_ML_DSA_65;
+  } else if (strcmp(name, "ML-DSA-87") == 0) {
+    nid = EVP_PKEY_ML_DSA_87;
+  } else if (strcmp(name, "ML-KEM-512") == 0) {
+    nid = EVP_PKEY_ML_KEM_512;
+  } else if (strcmp(name, "ML-KEM-768") == 0) {
+    nid = EVP_PKEY_ML_KEM_768;
+  } else if (strcmp(name, "ML-KEM-1024") == 0) {
+    nid = EVP_PKEY_ML_KEM_1024;
+#endif
+  } else {
+    nid = NID_undef;
   }
+  return nid;
 }
 }  // namespace
 
-Maybe<void> KeyObjectData::ToEncodedPublicKey(
+bool KeyObjectData::ToEncodedPublicKey(
     Environment* env,
-    const ncrypto::EVPKeyPointer::PublicKeyEncodingConfig& config,
+    const EVPKeyPointer::PublicKeyEncodingConfig& config,
     Local<Value>* out) {
   CHECK(key_type_ != KeyType::kKeyTypeSecret);
   if (config.output_key_object) {
     // Note that this has the downside of containing sensitive data of the
     // private key.
-    return NothingIfFalse(
-        KeyObjectHandle::Create(env, addRefWithType(KeyType::kKeyTypePublic))
-            .ToLocal(out));
-  } else if (config.format == ncrypto::EVPKeyPointer::PKFormatType::JWK) {
+    return KeyObjectHandle::Create(env, addRefWithType(KeyType::kKeyTypePublic))
+        .ToLocal(out);
+  } else if (config.format == EVPKeyPointer::PKFormatType::JWK) {
     *out = Object::New(env->isolate());
     return ExportJWKInner(
         env, addRefWithType(KeyType::kKeyTypePublic), *out, false);
   }
 
-  return NothingIfFalse(
-      WritePublicKey(env, GetAsymmetricKey(), config).ToLocal(out));
+  return WritePublicKey(env, GetAsymmetricKey(), config).ToLocal(out);
 }
 
-Maybe<void> KeyObjectData::ToEncodedPrivateKey(
+bool KeyObjectData::ToEncodedPrivateKey(
     Environment* env,
-    const ncrypto::EVPKeyPointer::PrivateKeyEncodingConfig& config,
+    const EVPKeyPointer::PrivateKeyEncodingConfig& config,
     Local<Value>* out) {
   CHECK(key_type_ != KeyType::kKeyTypeSecret);
   if (config.output_key_object) {
-    return NothingIfFalse(
-        KeyObjectHandle::Create(env, addRefWithType(KeyType::kKeyTypePrivate))
-            .ToLocal(out));
-  } else if (config.format == ncrypto::EVPKeyPointer::PKFormatType::JWK) {
+    return KeyObjectHandle::Create(env,
+                                   addRefWithType(KeyType::kKeyTypePrivate))
+        .ToLocal(out);
+  } else if (config.format == EVPKeyPointer::PKFormatType::JWK) {
     *out = Object::New(env->isolate());
     return ExportJWKInner(
         env, addRefWithType(KeyType::kKeyTypePrivate), *out, false);
   }
 
-  return NothingIfFalse(
-      WritePrivateKey(env, GetAsymmetricKey(), config).ToLocal(out));
+  return WritePrivateKey(env, GetAsymmetricKey(), config).ToLocal(out);
 }
 
-Maybe<ncrypto::EVPKeyPointer::PrivateKeyEncodingConfig>
+Maybe<EVPKeyPointer::PrivateKeyEncodingConfig>
 KeyObjectData::GetPrivateKeyEncodingFromJs(
     const FunctionCallbackInfo<Value>& args,
     unsigned int* offset,
     KeyEncodingContext context) {
   Environment* env = Environment::GetCurrent(args);
 
-  ncrypto::EVPKeyPointer::PrivateKeyEncodingConfig config;
+  EVPKeyPointer::PrivateKeyEncodingConfig config;
   if (!GetKeyFormatAndTypeFromJs(args, offset, context).To(&config)) {
-    return Nothing<ncrypto::EVPKeyPointer::PrivateKeyEncodingConfig>();
+    return Nothing<EVPKeyPointer::PrivateKeyEncodingConfig>();
   }
 
   if (config.output_key_object) {
@@ -343,10 +358,10 @@ KeyObjectData::GetPrivateKeyEncodingFromJs(
     if (context != kKeyContextInput) {
       if (args[*offset]->IsString()) {
         Utf8Value cipher_name(env->isolate(), args[*offset]);
-        config.cipher = EVP_get_cipherbyname(*cipher_name);
+        config.cipher = ncrypto::getCipherByName(*cipher_name);
         if (config.cipher == nullptr) {
           THROW_ERR_CRYPTO_UNKNOWN_CIPHER(env);
-          return Nothing<ncrypto::EVPKeyPointer::PrivateKeyEncodingConfig>();
+          return Nothing<EVPKeyPointer::PrivateKeyEncodingConfig>();
         }
         needs_passphrase = true;
       } else {
@@ -361,7 +376,7 @@ KeyObjectData::GetPrivateKeyEncodingFromJs(
       ArrayBufferOrViewContents<char> passphrase(args[*offset]);
       if (!passphrase.CheckSizeInt32()) [[unlikely]] {
         THROW_ERR_OUT_OF_RANGE(env, "passphrase is too big");
-        return Nothing<ncrypto::EVPKeyPointer::PrivateKeyEncodingConfig>();
+        return Nothing<EVPKeyPointer::PrivateKeyEncodingConfig>();
       }
       config.passphrase = passphrase.ToDataPointer();
     } else {
@@ -370,11 +385,10 @@ KeyObjectData::GetPrivateKeyEncodingFromJs(
   }
 
   (*offset)++;
-  return Just<ncrypto::EVPKeyPointer::PrivateKeyEncodingConfig>(
-      std::move(config));
+  return Just<EVPKeyPointer::PrivateKeyEncodingConfig>(std::move(config));
 }
 
-Maybe<ncrypto::EVPKeyPointer::PublicKeyEncodingConfig>
+Maybe<EVPKeyPointer::PublicKeyEncodingConfig>
 KeyObjectData::GetPublicKeyEncodingFromJs(
     const FunctionCallbackInfo<Value>& args,
     unsigned int* offset,
@@ -390,7 +404,7 @@ KeyObjectData KeyObjectData::GetPrivateKeyFromJs(
     Environment* env = Environment::GetCurrent(args);
     auto key = ByteSource::FromStringOrBuffer(env, args[(*offset)++]);
 
-    ncrypto::EVPKeyPointer::PrivateKeyEncodingConfig config;
+    EVPKeyPointer::PrivateKeyEncodingConfig config;
     if (!GetPrivateKeyEncodingFromJs(args, offset, kKeyContextInput)
              .To(&config)) {
       return {};
@@ -423,7 +437,7 @@ KeyObjectData KeyObjectData::GetPublicOrPrivateKeyFromJs(
       return {};
     }
 
-    ncrypto::EVPKeyPointer::PrivateKeyEncodingConfig config;
+    EVPKeyPointer::PrivateKeyEncodingConfig config;
     if (!KeyObjectData::GetPrivateKeyEncodingFromJs(
              args, offset, kKeyContextInput)
              .To(&config)) {
@@ -435,7 +449,7 @@ KeyObjectData KeyObjectData::GetPublicOrPrivateKeyFromJs(
         .len = data.size(),
     };
 
-    if (config.format == ncrypto::EVPKeyPointer::PKFormatType::PEM) {
+    if (config.format == EVPKeyPointer::PKFormatType::PEM) {
       // For PEM, we can easily determine whether it is a public or private key
       // by looking for the respective PEM tags.
       auto res = EVPKeyPointer::TryParsePublicKeyPEM(buffer);
@@ -456,13 +470,13 @@ KeyObjectData KeyObjectData::GetPublicOrPrivateKeyFromJs(
     static const auto is_public = [](const auto& config,
                                      const auto& buffer) -> bool {
       switch (config.type) {
-        case ncrypto::EVPKeyPointer::PKEncodingType::PKCS1:
+        case EVPKeyPointer::PKEncodingType::PKCS1:
           return !EVPKeyPointer::IsRSAPrivateKey(buffer);
-        case ncrypto::EVPKeyPointer::PKEncodingType::SPKI:
+        case EVPKeyPointer::PKEncodingType::SPKI:
           return true;
-        case ncrypto::EVPKeyPointer::PKEncodingType::PKCS8:
+        case EVPKeyPointer::PKEncodingType::PKCS8:
           return false;
-        case ncrypto::EVPKeyPointer::PKEncodingType::SEC1:
+        case EVPKeyPointer::PKEncodingType::SEC1:
           return false;
         default:
           UNREACHABLE("Invalid key encoding type");
@@ -497,6 +511,7 @@ KeyObjectData KeyObjectData::GetParsedKey(KeyType type,
                                           EVPKeyPointer&& pkey,
                                           ParseKeyResult ret,
                                           const char* default_msg) {
+  MarkPopErrorOnReturn mark_pop_error_on_return;
   switch (ret) {
     case ParseKeyResult::kParseKeyOk: {
       return CreateAsymmetric(type, std::move(pkey));
@@ -507,7 +522,7 @@ KeyObjectData KeyObjectData::GetParsedKey(KeyType type,
       return {};
     }
     default: {
-      ThrowCryptoError(env, ERR_get_error(), default_msg);
+      ThrowCryptoError(env, mark_pop_error_on_return.peekError(), default_msg);
       return {};
     }
   }
@@ -588,12 +603,12 @@ size_t KeyObjectData::GetSymmetricKeySize() const {
 }
 
 bool KeyObjectHandle::HasInstance(Environment* env, Local<Value> value) {
-  Local<FunctionTemplate> t = env->crypto_key_object_handle_constructor();
+  auto t = env->crypto_key_object_handle_constructor();
   return !t.IsEmpty() && t->HasInstance(value);
 }
 
-v8::Local<v8::Function> KeyObjectHandle::Initialize(Environment* env) {
-  Local<FunctionTemplate> templ = env->crypto_key_object_handle_constructor();
+Local<Function> KeyObjectHandle::Initialize(Environment* env) {
+  auto templ = env->crypto_key_object_handle_constructor();
   if (templ.IsEmpty()) {
     Isolate* isolate = env->isolate();
     templ = NewFunctionTemplate(isolate, New);
@@ -611,6 +626,11 @@ v8::Local<v8::Function> KeyObjectHandle::Initialize(Environment* env) {
     SetProtoMethod(isolate, templ, "exportJwk", ExportJWK);
     SetProtoMethod(isolate, templ, "initECRaw", InitECRaw);
     SetProtoMethod(isolate, templ, "initEDRaw", InitEDRaw);
+#if OPENSSL_WITH_PQC
+    SetProtoMethod(isolate, templ, "initPqcRaw", InitPqcRaw);
+    SetProtoMethodNoSideEffect(isolate, templ, "rawPublicKey", RawPublicKey);
+    SetProtoMethodNoSideEffect(isolate, templ, "rawSeed", RawSeed);
+#endif
     SetProtoMethod(isolate, templ, "initJwk", InitJWK);
     SetProtoMethod(isolate, templ, "keyDetail", GetKeyDetail);
     SetProtoMethod(isolate, templ, "equals", Equals);
@@ -631,6 +651,11 @@ void KeyObjectHandle::RegisterExternalReferences(
   registry->Register(ExportJWK);
   registry->Register(InitECRaw);
   registry->Register(InitEDRaw);
+#if OPENSSL_WITH_PQC
+  registry->Register(InitPqcRaw);
+  registry->Register(RawPublicKey);
+  registry->Register(RawSeed);
+#endif
   registry->Register(InitJWK);
   registry->Register(GetKeyDetail);
   registry->Register(Equals);
@@ -641,8 +666,9 @@ MaybeLocal<Object> KeyObjectHandle::Create(Environment* env,
   Local<Object> obj;
   Local<Function> ctor = KeyObjectHandle::Initialize(env);
   CHECK(!env->crypto_key_object_handle_constructor().IsEmpty());
-  if (!ctor->NewInstance(env->context(), 0, nullptr).ToLocal(&obj))
-    return MaybeLocal<Object>();
+  if (!ctor->NewInstance(env->context(), 0, nullptr).ToLocal(&obj)) {
+    return {};
+  }
 
   KeyObjectHandle* key = Unwrap<KeyObjectHandle>(obj);
   CHECK_NOT_NULL(key);
@@ -755,22 +781,21 @@ void KeyObjectHandle::InitECRaw(const FunctionCallbackInfo<Value>& args) {
   MarkPopErrorOnReturn mark_pop_error_on_return;
 
   int id = OBJ_txt2nid(*name);
-  ECKeyPointer eckey(EC_KEY_new_by_curve_name(id));
+  auto eckey = ECKeyPointer::NewByCurveName(id);
   if (!eckey)
     return args.GetReturnValue().Set(false);
 
-  const EC_GROUP* group = EC_KEY_get0_group(eckey.get());
-  ECPointPointer pub(ECDH::BufferToPoint(env, group, args[1]));
+  const auto group = eckey.getGroup();
+  auto pub = ECDH::BufferToPoint(env, group, args[1]);
 
-  if (!pub ||
-      !eckey ||
-      !EC_KEY_set_public_key(eckey.get(), pub.get())) {
+  if (!pub || !eckey || !eckey.setPublicKey(pub)) {
     return args.GetReturnValue().Set(false);
   }
 
   auto pkey = EVPKeyPointer::New();
-  if (!EVP_PKEY_assign_EC_KEY(pkey.get(), eckey.get()))
+  if (!pkey.assign(eckey)) {
     args.GetReturnValue().Set(false);
+  }
 
   eckey.release();  // Release ownership of the key
 
@@ -779,22 +804,6 @@ void KeyObjectHandle::InitECRaw(const FunctionCallbackInfo<Value>& args) {
   args.GetReturnValue().Set(true);
 }
 
-int GetOKPCurveFromName(const char* name) {
-  int nid;
-  if (strcmp(name, "Ed25519") == 0) {
-    nid = EVP_PKEY_ED25519;
-  } else if (strcmp(name, "Ed448") == 0) {
-    nid = EVP_PKEY_ED448;
-  } else if (strcmp(name, "X25519") == 0) {
-    nid = EVP_PKEY_X25519;
-  } else if (strcmp(name, "X448") == 0) {
-    nid = EVP_PKEY_X448;
-  } else {
-    nid = NID_undef;
-  }
-  return nid;
-}
-
 void KeyObjectHandle::InitEDRaw(const FunctionCallbackInfo<Value>& args) {
   Environment* env = Environment::GetCurrent(args);
   KeyObjectHandle* key;
@@ -804,7 +813,7 @@ void KeyObjectHandle::InitEDRaw(const FunctionCallbackInfo<Value>& args) {
   Utf8Value name(env->isolate(), args[0]);
 
   ArrayBufferOrViewContents<unsigned char> key_data(args[1]);
-  KeyType type = static_cast<KeyType>(args[2].As<Int32>()->Value());
+  KeyType type = FromV8Value<KeyType>(args[2]);
 
   MarkPopErrorOnReturn mark_pop_error_on_return;
 
@@ -813,7 +822,7 @@ void KeyObjectHandle::InitEDRaw(const FunctionCallbackInfo<Value>& args) {
   new_key_fn fn = type == kKeyTypePrivate ? EVPKeyPointer::NewRawPrivate
                                           : EVPKeyPointer::NewRawPublic;
 
-  int id = GetOKPCurveFromName(*name);
+  int id = GetNidFromName(*name);
 
   switch (id) {
     case EVP_PKEY_X25519:
@@ -839,6 +848,54 @@ void KeyObjectHandle::InitEDRaw(const FunctionCallbackInfo<Value>& args) {
   args.GetReturnValue().Set(true);
 }
 
+#if OPENSSL_WITH_PQC
+void KeyObjectHandle::InitPqcRaw(const FunctionCallbackInfo<Value>& args) {
+  Environment* env = Environment::GetCurrent(args);
+  KeyObjectHandle* key;
+  ASSIGN_OR_RETURN_UNWRAP(&key, args.This());
+
+  CHECK(args[0]->IsString());
+  Utf8Value name(env->isolate(), args[0]);
+
+  ArrayBufferOrViewContents<unsigned char> key_data(args[1]);
+  KeyType type = FromV8Value<KeyType>(args[2]);
+
+  MarkPopErrorOnReturn mark_pop_error_on_return;
+
+  typedef EVPKeyPointer (*new_key_fn)(
+      int, const ncrypto::Buffer<const unsigned char>&);
+  new_key_fn fn = type == kKeyTypePrivate ? EVPKeyPointer::NewRawSeed
+                                          : EVPKeyPointer::NewRawPublic;
+
+  int id = GetNidFromName(*name);
+
+  switch (id) {
+    case EVP_PKEY_ML_DSA_44:
+    case EVP_PKEY_ML_DSA_65:
+    case EVP_PKEY_ML_DSA_87:
+    case EVP_PKEY_ML_KEM_512:
+    case EVP_PKEY_ML_KEM_768:
+    case EVP_PKEY_ML_KEM_1024: {
+      auto pkey = fn(id,
+                     ncrypto::Buffer<const unsigned char>{
+                         .data = key_data.data(),
+                         .len = key_data.size(),
+                     });
+      if (!pkey) {
+        return args.GetReturnValue().Set(false);
+      }
+      key->data_ = KeyObjectData::CreateAsymmetric(type, std::move(pkey));
+      CHECK(key->data_);
+      break;
+    }
+    default:
+      UNREACHABLE();
+  }
+
+  args.GetReturnValue().Set(true);
+}
+#endif
+
 void KeyObjectHandle::Equals(const FunctionCallbackInfo<Value>& args) {
   KeyObjectHandle* self_handle;
   KeyObjectHandle* arg_handle;
@@ -863,6 +920,7 @@ void KeyObjectHandle::Equals(const FunctionCallbackInfo<Value>& args) {
       break;
     }
     case kKeyTypePublic:
+      // Fall through
     case kKeyTypePrivate: {
       EVP_PKEY* pkey = key.GetAsymmetricKey().get();
       EVP_PKEY* pkey2 = key2.GetAsymmetricKey().get();
@@ -894,22 +952,16 @@ void KeyObjectHandle::GetKeyDetail(const FunctionCallbackInfo<Value>& args) {
 
   const auto& data = key->Data();
 
-  switch (data.GetKeyType()) {
-    case kKeyTypeSecret:
-      if (GetSecretKeyDetail(env, data, args[0].As<Object>()).IsNothing())
-        return;
-      break;
-    case kKeyTypePublic:
-      // Fall through
-    case kKeyTypePrivate:
-      if (GetAsymmetricKeyDetail(env, data, args[0].As<Object>()).IsNothing())
-        return;
-      break;
-    default:
-      UNREACHABLE();
+  if (data.GetKeyType() == kKeyTypeSecret) {
+    if (GetSecretKeyDetail(env, data, args[0].As<Object>())) [[likely]] {
+      args.GetReturnValue().Set(args[0]);
+    }
+    return;
   }
 
-  args.GetReturnValue().Set(args[0]);
+  if (GetAsymmetricKeyDetail(env, data, args[0].As<Object>())) [[likely]] {
+    args.GetReturnValue().Set(args[0]);
+  }
 }
 
 Local<Value> KeyObjectHandle::GetAsymmetricKeyType() const {
@@ -932,6 +984,44 @@ Local<Value> KeyObjectHandle::GetAsymmetricKeyType() const {
       return env()->crypto_x25519_string();
     case EVP_PKEY_X448:
       return env()->crypto_x448_string();
+#if OPENSSL_WITH_PQC
+    case EVP_PKEY_ML_DSA_44:
+      return env()->crypto_ml_dsa_44_string();
+    case EVP_PKEY_ML_DSA_65:
+      return env()->crypto_ml_dsa_65_string();
+    case EVP_PKEY_ML_DSA_87:
+      return env()->crypto_ml_dsa_87_string();
+    case EVP_PKEY_ML_KEM_512:
+      return env()->crypto_ml_kem_512_string();
+    case EVP_PKEY_ML_KEM_768:
+      return env()->crypto_ml_kem_768_string();
+    case EVP_PKEY_ML_KEM_1024:
+      return env()->crypto_ml_kem_1024_string();
+    case EVP_PKEY_SLH_DSA_SHA2_128F:
+      return env()->crypto_slh_dsa_sha2_128f_string();
+    case EVP_PKEY_SLH_DSA_SHA2_128S:
+      return env()->crypto_slh_dsa_sha2_128s_string();
+    case EVP_PKEY_SLH_DSA_SHA2_192F:
+      return env()->crypto_slh_dsa_sha2_192f_string();
+    case EVP_PKEY_SLH_DSA_SHA2_192S:
+      return env()->crypto_slh_dsa_sha2_192s_string();
+    case EVP_PKEY_SLH_DSA_SHA2_256F:
+      return env()->crypto_slh_dsa_sha2_256f_string();
+    case EVP_PKEY_SLH_DSA_SHA2_256S:
+      return env()->crypto_slh_dsa_sha2_256s_string();
+    case EVP_PKEY_SLH_DSA_SHAKE_128F:
+      return env()->crypto_slh_dsa_shake_128f_string();
+    case EVP_PKEY_SLH_DSA_SHAKE_128S:
+      return env()->crypto_slh_dsa_shake_128s_string();
+    case EVP_PKEY_SLH_DSA_SHAKE_192F:
+      return env()->crypto_slh_dsa_shake_192f_string();
+    case EVP_PKEY_SLH_DSA_SHAKE_192S:
+      return env()->crypto_slh_dsa_shake_192s_string();
+    case EVP_PKEY_SLH_DSA_SHAKE_256F:
+      return env()->crypto_slh_dsa_shake_256f_string();
+    case EVP_PKEY_SLH_DSA_SHAKE_256S:
+      return env()->crypto_slh_dsa_shake_256s_string();
+#endif
     default:
       return Undefined(env()->isolate());
   }
@@ -953,15 +1043,8 @@ bool KeyObjectHandle::CheckEcKeyData() const {
   CHECK(ctx);
   CHECK_EQ(key.id(), EVP_PKEY_EC);
 
-  if (data_.GetKeyType() == kKeyTypePrivate) {
-    return EVP_PKEY_check(ctx.get()) == 1;
-  }
-
-#if OPENSSL_VERSION_MAJOR >= 3
-  return EVP_PKEY_public_check_quick(ctx.get()) == 1;
-#else
-  return EVP_PKEY_public_check(ctx.get()) == 1;
-#endif
+  return data_.GetKeyType() == kKeyTypePrivate ? ctx.privateCheck()
+                                               : ctx.publicCheck();
 }
 
 void KeyObjectHandle::CheckEcKeyData(const FunctionCallbackInfo<Value>& args) {
@@ -984,53 +1067,103 @@ void KeyObjectHandle::Export(const FunctionCallbackInfo<Value>& args) {
   ASSIGN_OR_RETURN_UNWRAP(&key, args.This());
 
   KeyType type = key->Data().GetKeyType();
+  unsigned int offset = 0;
 
-  MaybeLocal<Value> result;
+  Local<Value> result;
   if (type == kKeyTypeSecret) {
-    result = key->ExportSecretKey();
-  } else if (type == kKeyTypePublic) {
-    unsigned int offset = 0;
-    ncrypto::EVPKeyPointer::PublicKeyEncodingConfig config;
+    if (key->ExportSecretKey().ToLocal(&result)) [[likely]] {
+      args.GetReturnValue().Set(result);
+    }
+    return;
+  }
+
+  if (type == kKeyTypePublic) {
+    EVPKeyPointer::PublicKeyEncodingConfig config;
     if (!KeyObjectData::GetPublicKeyEncodingFromJs(
              args, &offset, kKeyContextExport)
              .To(&config)) {
       return;
     }
     CHECK_EQ(offset, static_cast<unsigned int>(args.Length()));
-    result = key->ExportPublicKey(config);
-  } else {
-    CHECK_EQ(type, kKeyTypePrivate);
-    unsigned int offset = 0;
-    ncrypto::EVPKeyPointer::PrivateKeyEncodingConfig config;
-    if (!KeyObjectData::GetPrivateKeyEncodingFromJs(
-             args, &offset, kKeyContextExport)
-             .To(&config)) {
-      return;
+    if (key->ExportPublicKey(config).ToLocal(&result)) [[likely]] {
+      args.GetReturnValue().Set(result);
     }
-    CHECK_EQ(offset, static_cast<unsigned int>(args.Length()));
-    result = key->ExportPrivateKey(config);
+    return;
   }
 
-  if (!result.IsEmpty())
-    args.GetReturnValue().Set(result.FromMaybe(Local<Value>()));
+  CHECK_EQ(type, kKeyTypePrivate);
+  EVPKeyPointer::PrivateKeyEncodingConfig config;
+  if (!KeyObjectData::GetPrivateKeyEncodingFromJs(
+           args, &offset, kKeyContextExport)
+           .To(&config)) {
+    return;
+  }
+  CHECK_EQ(offset, static_cast<unsigned int>(args.Length()));
+  if (key->ExportPrivateKey(config).ToLocal(&result)) [[likely]] {
+    args.GetReturnValue().Set(result);
+  }
 }
 
 MaybeLocal<Value> KeyObjectHandle::ExportSecretKey() const {
-  const char* buf = data_.GetSymmetricKey();
-  unsigned int len = data_.GetSymmetricKeySize();
-  return Buffer::Copy(env(), buf, len).FromMaybe(Local<Value>());
+  return Buffer::Copy(
+             env(), data_.GetSymmetricKey(), data_.GetSymmetricKeySize())
+      .FromMaybe(Local<Value>());
 }
 
 MaybeLocal<Value> KeyObjectHandle::ExportPublicKey(
-    const ncrypto::EVPKeyPointer::PublicKeyEncodingConfig& config) const {
+    const EVPKeyPointer::PublicKeyEncodingConfig& config) const {
   return WritePublicKey(env(), data_.GetAsymmetricKey(), config);
 }
 
 MaybeLocal<Value> KeyObjectHandle::ExportPrivateKey(
-    const ncrypto::EVPKeyPointer::PrivateKeyEncodingConfig& config) const {
+    const EVPKeyPointer::PrivateKeyEncodingConfig& config) const {
   return WritePrivateKey(env(), data_.GetAsymmetricKey(), config);
 }
 
+#if OPENSSL_WITH_PQC
+void KeyObjectHandle::RawPublicKey(
+    const v8::FunctionCallbackInfo<v8::Value>& args) {
+  Environment* env = Environment::GetCurrent(args);
+  KeyObjectHandle* key;
+  ASSIGN_OR_RETURN_UNWRAP(&key, args.This());
+
+  const KeyObjectData& data = key->Data();
+  CHECK_NE(data.GetKeyType(), kKeyTypeSecret);
+
+  Mutex::ScopedLock lock(data.mutex());
+  auto raw_data = data.GetAsymmetricKey().rawPublicKey();
+  if (!raw_data) {
+    return THROW_ERR_CRYPTO_OPERATION_FAILED(env,
+                                             "Failed to get raw public key");
+  }
+
+  args.GetReturnValue().Set(
+      Buffer::Copy(
+          env, reinterpret_cast<const char*>(raw_data.get()), raw_data.size())
+          .FromMaybe(Local<Value>()));
+}
+
+void KeyObjectHandle::RawSeed(const v8::FunctionCallbackInfo<v8::Value>& args) {
+  Environment* env = Environment::GetCurrent(args);
+  KeyObjectHandle* key;
+  ASSIGN_OR_RETURN_UNWRAP(&key, args.This());
+
+  const KeyObjectData& data = key->Data();
+  CHECK_EQ(data.GetKeyType(), kKeyTypePrivate);
+
+  Mutex::ScopedLock lock(data.mutex());
+  auto raw_data = data.GetAsymmetricKey().rawSeed();
+  if (!raw_data) {
+    return THROW_ERR_CRYPTO_OPERATION_FAILED(env, "Failed to get raw seed");
+  }
+
+  args.GetReturnValue().Set(
+      Buffer::Copy(
+          env, reinterpret_cast<const char*>(raw_data.get()), raw_data.size())
+          .FromMaybe(Local<Value>()));
+}
+#endif
+
 void KeyObjectHandle::ExportJWK(
     const v8::FunctionCallbackInfo<v8::Value>& args) {
   Environment* env = Environment::GetCurrent(args);
@@ -1040,9 +1173,9 @@ void KeyObjectHandle::ExportJWK(
   CHECK(args[0]->IsObject());
   CHECK(args[1]->IsBoolean());
 
-  ExportJWKInner(env, key->Data(), args[0], args[1]->IsTrue());
-
-  args.GetReturnValue().Set(args[0]);
+  if (ExportJWKInner(env, key->Data(), args[0], args[1]->IsTrue())) {
+    args.GetReturnValue().Set(args[0]);
+  }
 }
 
 void NativeKeyObject::Initialize(Environment* env, Local<Object> target) {
@@ -1063,6 +1196,7 @@ void NativeKeyObject::New(const FunctionCallbackInfo<Value>& args) {
   CHECK_EQ(args.Length(), 1);
   CHECK(args[0]->IsObject());
   KeyObjectHandle* handle = Unwrap<KeyObjectHandle>(args[0].As<Object>());
+  CHECK_NOT_NULL(handle);
   new NativeKeyObject(env, args.This(), handle->Data());
 }
 
@@ -1184,19 +1318,22 @@ void Initialize(Environment* env, Local<Object> target) {
               KeyObjectHandle::Initialize(env)).Check();
 
   constexpr int kKeyEncodingPKCS1 =
-      static_cast<int>(ncrypto::EVPKeyPointer::PKEncodingType::PKCS1);
+      static_cast<int>(EVPKeyPointer::PKEncodingType::PKCS1);
   constexpr int kKeyEncodingPKCS8 =
-      static_cast<int>(ncrypto::EVPKeyPointer::PKEncodingType::PKCS8);
+      static_cast<int>(EVPKeyPointer::PKEncodingType::PKCS8);
   constexpr int kKeyEncodingSPKI =
-      static_cast<int>(ncrypto::EVPKeyPointer::PKEncodingType::SPKI);
+      static_cast<int>(EVPKeyPointer::PKEncodingType::SPKI);
   constexpr int kKeyEncodingSEC1 =
-      static_cast<int>(ncrypto::EVPKeyPointer::PKEncodingType::SEC1);
+      static_cast<int>(EVPKeyPointer::PKEncodingType::SEC1);
   constexpr int kKeyFormatDER =
-      static_cast<int>(ncrypto::EVPKeyPointer::PKFormatType::DER);
+      static_cast<int>(EVPKeyPointer::PKFormatType::DER);
   constexpr int kKeyFormatPEM =
-      static_cast<int>(ncrypto::EVPKeyPointer::PKFormatType::PEM);
+      static_cast<int>(EVPKeyPointer::PKFormatType::PEM);
   constexpr int kKeyFormatJWK =
-      static_cast<int>(ncrypto::EVPKeyPointer::PKFormatType::JWK);
+      static_cast<int>(EVPKeyPointer::PKFormatType::JWK);
+
+  constexpr auto kSigEncDER = DSASigEnc::DER;
+  constexpr auto kSigEncP1363 = DSASigEnc::P1363;
 
   NODE_DEFINE_CONSTANT(target, kWebCryptoKeyFormatRaw);
   NODE_DEFINE_CONSTANT(target, kWebCryptoKeyFormatPKCS8);
@@ -1204,6 +1341,26 @@ void Initialize(Environment* env, Local<Object> target) {
   NODE_DEFINE_CONSTANT(target, kWebCryptoKeyFormatJWK);
   NODE_DEFINE_CONSTANT(target, EVP_PKEY_ED25519);
   NODE_DEFINE_CONSTANT(target, EVP_PKEY_ED448);
+#if OPENSSL_WITH_PQC
+  NODE_DEFINE_CONSTANT(target, EVP_PKEY_ML_DSA_44);
+  NODE_DEFINE_CONSTANT(target, EVP_PKEY_ML_DSA_65);
+  NODE_DEFINE_CONSTANT(target, EVP_PKEY_ML_DSA_87);
+  NODE_DEFINE_CONSTANT(target, EVP_PKEY_ML_KEM_512);
+  NODE_DEFINE_CONSTANT(target, EVP_PKEY_ML_KEM_768);
+  NODE_DEFINE_CONSTANT(target, EVP_PKEY_ML_KEM_1024);
+  NODE_DEFINE_CONSTANT(target, EVP_PKEY_SLH_DSA_SHA2_128F);
+  NODE_DEFINE_CONSTANT(target, EVP_PKEY_SLH_DSA_SHA2_128S);
+  NODE_DEFINE_CONSTANT(target, EVP_PKEY_SLH_DSA_SHA2_192F);
+  NODE_DEFINE_CONSTANT(target, EVP_PKEY_SLH_DSA_SHA2_192S);
+  NODE_DEFINE_CONSTANT(target, EVP_PKEY_SLH_DSA_SHA2_256F);
+  NODE_DEFINE_CONSTANT(target, EVP_PKEY_SLH_DSA_SHA2_256S);
+  NODE_DEFINE_CONSTANT(target, EVP_PKEY_SLH_DSA_SHAKE_128F);
+  NODE_DEFINE_CONSTANT(target, EVP_PKEY_SLH_DSA_SHAKE_128S);
+  NODE_DEFINE_CONSTANT(target, EVP_PKEY_SLH_DSA_SHAKE_192F);
+  NODE_DEFINE_CONSTANT(target, EVP_PKEY_SLH_DSA_SHAKE_192S);
+  NODE_DEFINE_CONSTANT(target, EVP_PKEY_SLH_DSA_SHAKE_256F);
+  NODE_DEFINE_CONSTANT(target, EVP_PKEY_SLH_DSA_SHAKE_256S);
+#endif
   NODE_DEFINE_CONSTANT(target, EVP_PKEY_X25519);
   NODE_DEFINE_CONSTANT(target, EVP_PKEY_X448);
   NODE_DEFINE_CONSTANT(target, kKeyEncodingPKCS1);
