@@ -20,6 +20,13 @@
 
 namespace node {
 
+using ncrypto::BIOPointer;
+using ncrypto::ClearErrorOnReturn;
+using ncrypto::MarkPopErrorOnReturn;
+using ncrypto::SSLCtxPointer;
+using ncrypto::SSLPointer;
+using ncrypto::SSLSessionPointer;
+using ncrypto::X509Pointer;
 using v8::ArrayBuffer;
 using v8::Just;
 using v8::Local;
@@ -27,6 +34,7 @@ using v8::Maybe;
 using v8::MaybeLocal;
 using v8::Nothing;
 using v8::Object;
+using v8::Undefined;
 using v8::Value;
 
 namespace quic {
@@ -43,7 +51,7 @@ namespace {
 //   return 0;
 // }();
 
-void EnableTrace(Environment* env, crypto::BIOPointer* bio, SSL* ssl) {
+void EnableTrace(Environment* env, BIOPointer* bio, SSL* ssl) {
 #if HAVE_SSL_TRACE
   static bool warn_trace_tls = true;
   if (warn_trace_tls) {
@@ -63,7 +71,7 @@ void EnableTrace(Environment* env, crypto::BIOPointer* bio, SSL* ssl) {
            size_t len,
            SSL* ssl,
            void* arg) -> void {
-          crypto::MarkPopErrorOnReturn mark_pop_error_on_return;
+          MarkPopErrorOnReturn mark_pop_error_on_return;
           SSL_trace(write_p, version, content_type, buf, len, ssl, arg);
         });
     SSL_set_msg_callback_arg(ssl, bio->get());
@@ -74,9 +82,9 @@ void EnableTrace(Environment* env, crypto::BIOPointer* bio, SSL* ssl) {
 template <typename T, typename Opt, std::vector<T> Opt::*member>
 bool SetOption(Environment* env,
                Opt* options,
-               const v8::Local<v8::Object>& object,
-               const v8::Local<v8::String>& name) {
-  v8::Local<v8::Value> value;
+               const Local<Object>& object,
+               const Local<v8::String>& name) {
+  Local<Value> value;
   if (!object->Get(env->context(), name).ToLocal(&value)) return false;
 
   if (value->IsUndefined()) return true;
@@ -88,7 +96,7 @@ bool SetOption(Environment* env,
     auto values = value.As<v8::Array>();
     uint32_t count = values->Length();
     for (uint32_t n = 0; n < count; n++) {
-      v8::Local<v8::Value> item;
+      Local<Value> item;
       if (!values->Get(context, n).ToLocal(&item)) {
         return false;
       }
@@ -107,7 +115,7 @@ bool SetOption(Environment* env,
         if (item->IsArrayBufferView()) {
           (options->*member).emplace_back(item.As<v8::ArrayBufferView>());
         } else if (item->IsArrayBuffer()) {
-          (options->*member).emplace_back(item.As<v8::ArrayBuffer>());
+          (options->*member).emplace_back(item.As<ArrayBuffer>());
         } else {
           Utf8Value namestr(env->isolate(), name);
           THROW_ERR_INVALID_ARG_TYPE(
@@ -132,7 +140,7 @@ bool SetOption(Environment* env,
       if (value->IsArrayBufferView()) {
         (options->*member).emplace_back(value.As<v8::ArrayBufferView>());
       } else if (value->IsArrayBuffer()) {
-        (options->*member).emplace_back(value.As<v8::ArrayBuffer>());
+        (options->*member).emplace_back(value.As<ArrayBuffer>());
       } else {
         Utf8Value namestr(env->isolate(), name);
         THROW_ERR_INVALID_ARG_TYPE(
@@ -170,7 +178,7 @@ int TLSContext::OnSelectAlpn(SSL* ssl,
   static constexpr size_t kMaxAlpnLen = 255;
   auto& session = TLSSession::From(ssl);
 
-  const auto& requested = session.context().options().alpn;
+  const auto& requested = session.context().options().protocol;
   if (requested.length() > kMaxAlpnLen) return SSL_TLSEXT_ERR_NOACK;
 
   // The Session supports exactly one ALPN identifier. If that does not match
@@ -240,12 +248,12 @@ std::unique_ptr<TLSSession> TLSContext::NewSession(
       session, shared_from_this(), maybeSessionTicket);
 }
 
-crypto::SSLCtxPointer TLSContext::Initialize() {
-  crypto::SSLCtxPointer ctx;
+SSLCtxPointer TLSContext::Initialize() {
+  SSLCtxPointer ctx;
   switch (side_) {
     case Side::SERVER: {
       static constexpr unsigned char kSidCtx[] = "Node.js QUIC Server";
-      ctx.reset(SSL_CTX_new(TLS_server_method()));
+      ctx = SSLCtxPointer::NewServer();
       CHECK_EQ(ngtcp2_crypto_quictls_configure_server_context(ctx.get()), 0);
       CHECK_EQ(SSL_CTX_set_max_early_data(ctx.get(), UINT32_MAX), 1);
       SSL_CTX_set_options(ctx.get(),
@@ -266,15 +274,17 @@ crypto::SSLCtxPointer TLSContext::Initialize() {
                            OnVerifyClientCertificate);
       }
 
-      CHECK_EQ(SSL_CTX_set_session_ticket_cb(ctx.get(),
-                                             SessionTicket::GenerateCallback,
-                                             SessionTicket::DecryptedCallback,
-                                             nullptr),
-               1);
+      // TODO(@jasnell): There's a bug int the GenerateCallback flow somewhere.
+      // Need to update in order to support session tickets.
+      // CHECK_EQ(SSL_CTX_set_session_ticket_cb(ctx.get(),
+      //                                        SessionTicket::GenerateCallback,
+      //                                        SessionTicket::DecryptedCallback,
+      //                                        nullptr),
+      //          1);
       break;
     }
     case Side::CLIENT: {
-      ctx_.reset(SSL_CTX_new(TLS_client_method()));
+      ctx = SSLCtxPointer::NewClient();
       CHECK_EQ(ngtcp2_crypto_quictls_configure_client_context(ctx.get()), 0);
 
       SSL_CTX_set_session_cache_mode(
@@ -289,16 +299,16 @@ crypto::SSLCtxPointer TLSContext::Initialize() {
 
   if (SSL_CTX_set_ciphersuites(ctx.get(), options_.ciphers.c_str()) != 1) {
     validation_error_ = "Invalid cipher suite";
-    return crypto::SSLCtxPointer();
+    return SSLCtxPointer();
   }
 
   if (SSL_CTX_set1_groups_list(ctx.get(), options_.groups.c_str()) != 1) {
     validation_error_ = "Invalid cipher groups";
-    return crypto::SSLCtxPointer();
+    return SSLCtxPointer();
   }
 
   {
-    crypto::ClearErrorOnReturn clear_error_on_return;
+    ClearErrorOnReturn clear_error_on_return;
     if (options_.ca.empty()) {
       auto store = crypto::GetOrCreateRootCertStore();
       X509_STORE_up_ref(store);
@@ -311,14 +321,12 @@ crypto::SSLCtxPointer TLSContext::Initialize() {
           X509_STORE_up_ref(store);
           SSL_CTX_set_cert_store(ctx.get(), store);
         } else {
-          crypto::BIOPointer bio = crypto::NodeBIO::NewFixed(buf.base, buf.len);
+          BIOPointer bio = crypto::NodeBIO::NewFixed(buf.base, buf.len);
           CHECK(bio);
           X509_STORE* cert_store = SSL_CTX_get_cert_store(ctx.get());
-          while (crypto::X509Pointer x509 = crypto::X509Pointer(
-                     PEM_read_bio_X509_AUX(bio.get(),
-                                           nullptr,
-                                           crypto::NoPasswordCallback,
-                                           nullptr))) {
+          while (
+              auto x509 = X509Pointer(PEM_read_bio_X509_AUX(
+                  bio.get(), nullptr, crypto::NoPasswordCallback, nullptr))) {
             if (cert_store == crypto::GetOrCreateRootCertStore()) {
               cert_store = crypto::NewRootCertStore();
               SSL_CTX_set_cert_store(ctx.get(), cert_store);
@@ -332,48 +340,48 @@ crypto::SSLCtxPointer TLSContext::Initialize() {
   }
 
   {
-    crypto::ClearErrorOnReturn clear_error_on_return;
+    ClearErrorOnReturn clear_error_on_return;
     for (const auto& cert : options_.certs) {
       uv_buf_t buf = cert;
       if (buf.len > 0) {
-        crypto::BIOPointer bio = crypto::NodeBIO::NewFixed(buf.base, buf.len);
+        BIOPointer bio = crypto::NodeBIO::NewFixed(buf.base, buf.len);
         CHECK(bio);
         cert_.reset();
         issuer_.reset();
         if (crypto::SSL_CTX_use_certificate_chain(
                 ctx.get(), std::move(bio), &cert_, &issuer_) == 0) {
           validation_error_ = "Invalid certificate";
-          return crypto::SSLCtxPointer();
+          return SSLCtxPointer();
         }
       }
     }
   }
 
   {
-    crypto::ClearErrorOnReturn clear_error_on_return;
+    ClearErrorOnReturn clear_error_on_return;
     for (const auto& key : options_.keys) {
       if (key.GetKeyType() != crypto::KeyType::kKeyTypePrivate) {
         validation_error_ = "Invalid key";
-        return crypto::SSLCtxPointer();
+        return SSLCtxPointer();
       }
       if (!SSL_CTX_use_PrivateKey(ctx.get(), key.GetAsymmetricKey().get())) {
         validation_error_ = "Invalid key";
-        return crypto::SSLCtxPointer();
+        return SSLCtxPointer();
       }
     }
   }
 
   {
-    crypto::ClearErrorOnReturn clear_error_on_return;
+    ClearErrorOnReturn clear_error_on_return;
     for (const auto& crl : options_.crl) {
       uv_buf_t buf = crl;
-      crypto::BIOPointer bio = crypto::NodeBIO::NewFixed(buf.base, buf.len);
+      BIOPointer bio = crypto::NodeBIO::NewFixed(buf.base, buf.len);
       DeleteFnPtr<X509_CRL, X509_CRL_free> crlptr(PEM_read_bio_X509_CRL(
           bio.get(), nullptr, crypto::NoPasswordCallback, nullptr));
 
       if (!crlptr) {
         validation_error_ = "Invalid CRL";
-        return crypto::SSLCtxPointer();
+        return SSLCtxPointer();
       }
 
       X509_STORE* cert_store = SSL_CTX_get_cert_store(ctx.get());
@@ -391,11 +399,11 @@ crypto::SSLCtxPointer TLSContext::Initialize() {
   }
 
   {
-    crypto::ClearErrorOnReturn clear_error_on_return;
+    ClearErrorOnReturn clear_error_on_return;
     if (options_.verify_private_key &&
         SSL_CTX_check_private_key(ctx.get()) != 1) {
       validation_error_ = "Invalid private key";
-      return crypto::SSLCtxPointer();
+      return SSLCtxPointer();
     }
   }
 
@@ -416,7 +424,7 @@ Maybe<TLSContext::Options> TLSContext::Options::From(Environment* env,
   auto& state = BindingData::Get(env);
 
   if (value->IsUndefined()) {
-    return Just(TLSContext::Options::kDefault);
+    return Just(kDefault);
   }
 
   if (!value->IsObject()) {
@@ -434,11 +442,11 @@ Maybe<TLSContext::Options> TLSContext::Options::From(Environment* env,
   SetOption<TLSContext::Options, &TLSContext::Options::name>(                  \
       env, &options, params, state.name##_string())
 
-  if (!SET(verify_client) || !SET(enable_tls_trace) || !SET(alpn) ||
-      !SET(sni) || !SET(ciphers) || !SET(groups) || !SET(verify_private_key) ||
-      !SET(keylog) || !SET_VECTOR(crypto::KeyObjectData, keys) ||
-      !SET_VECTOR(Store, certs) || !SET_VECTOR(Store, ca) ||
-      !SET_VECTOR(Store, crl)) {
+  if (!SET(verify_client) || !SET(enable_tls_trace) || !SET(protocol) ||
+      !SET(servername) || !SET(ciphers) || !SET(groups) ||
+      !SET(verify_private_key) || !SET(keylog) ||
+      !SET_VECTOR(crypto::KeyObjectData, keys) || !SET_VECTOR(Store, certs) ||
+      !SET_VECTOR(Store, ca) || !SET_VECTOR(Store, crl)) {
     return Nothing<Options>();
   }
 
@@ -449,8 +457,8 @@ std::string TLSContext::Options::ToString() const {
   DebugIndentScope indent;
   auto prefix = indent.Prefix();
   std::string res("{");
-  res += prefix + "alpn: " + alpn;
-  res += prefix + "sni: " + sni;
+  res += prefix + "protocol: " + protocol;
+  res += prefix + "servername: " + servername;
   res +=
       prefix + "keylog: " + (keylog ? std::string("yes") : std::string("no"));
   res += prefix + "verify client: " +
@@ -496,6 +504,12 @@ TLSSession::TLSSession(Session* session,
   Debug(session_, "Created new TLS session for %s", session->config().dcid);
 }
 
+TLSSession::~TLSSession() {
+  if (ssl_) {
+    SSL_set_app_data(ssl_.get(), nullptr);
+  }
+}
+
 TLSSession::operator SSL*() const {
   CHECK(ssl_);
   return ssl_.get();
@@ -506,17 +520,17 @@ bool TLSSession::early_data_was_accepted() const {
   return SSL_get_early_data_status(*this) == SSL_EARLY_DATA_ACCEPTED;
 }
 
-crypto::SSLPointer TLSSession::Initialize(
+SSLPointer TLSSession::Initialize(
     const std::optional<SessionTicket>& maybeSessionTicket) {
   auto& ctx = context();
   auto& options = ctx.options();
-  crypto::SSLPointer ssl(SSL_new(ctx));
+  SSLPointer ssl(SSL_new(ctx));
   SSL_set_app_data(ssl.get(), &ref_);
   ngtcp2_conn_set_tls_native_handle(*session_, ssl.get());
 
   // Enable tracing if the `--trace-tls` command line flag is used.
-  if (UNLIKELY(session_->env()->options()->trace_tls ||
-               options.enable_tls_trace)) {
+  if (session_->env()->options()->trace_tls || options.enable_tls_trace)
+      [[unlikely]] {
     EnableTrace(session_->env(), &bio_trace_, *this);
   }
 
@@ -530,14 +544,14 @@ crypto::SSLPointer TLSSession::Initialize(
       SSL_set_connect_state(ssl.get());
       if (SSL_set_alpn_protos(
               ssl.get(),
-              reinterpret_cast<const unsigned char*>(options.alpn.data()),
-              options.alpn.size()) != 0) {
+              reinterpret_cast<const unsigned char*>(options.protocol.data()),
+              options.protocol.size()) != 0) {
         validation_error_ = "Invalid ALPN";
-        return crypto::SSLPointer();
+        return SSLPointer();
       }
 
-      if (!options.sni.empty()) {
-        SSL_set_tlsext_host_name(ssl.get(), options.sni.data());
+      if (!options.servername.empty()) {
+        SSL_set_tlsext_host_name(ssl.get(), options.servername.data());
       } else {
         SSL_set_tlsext_host_name(ssl.get(), "localhost");
       }
@@ -545,11 +559,11 @@ crypto::SSLPointer TLSSession::Initialize(
       if (maybeSessionTicket.has_value()) {
         auto sessionTicket = maybeSessionTicket.value();
         uv_buf_t buf = sessionTicket.ticket();
-        crypto::SSLSessionPointer ticket = crypto::GetTLSSession(
+        SSLSessionPointer ticket = crypto::GetTLSSession(
             reinterpret_cast<unsigned char*>(buf.base), buf.len);
 
         // The early data will just be ignored if it's invalid.
-        if (crypto::SetTLSSession(ssl, ticket) &&
+        if (ssl.setSession(ticket) &&
             SSL_SESSION_get_max_early_data(ticket.get()) != 0) {
           ngtcp2_vec rtp = sessionTicket.transport_params();
           if (ngtcp2_conn_decode_and_set_0rtt_transport_params(
@@ -576,7 +590,7 @@ crypto::SSLPointer TLSSession::Initialize(
 
 std::optional<TLSSession::PeerIdentityValidationError>
 TLSSession::VerifyPeerIdentity(Environment* env) {
-  int err = crypto::VerifyPeerCertificate(ssl_);
+  int err = ssl_.verifyPeerCertificate().value_or(X509_V_ERR_UNSPECIFIED);
   if (err == X509_V_OK) return std::nullopt;
   Local<Value> reason;
   Local<Value> code;
@@ -606,20 +620,22 @@ MaybeLocal<Object> TLSSession::ephemeral_key(Environment* env) const {
 }
 
 MaybeLocal<Value> TLSSession::cipher_name(Environment* env) const {
-  return crypto::GetCurrentCipherName(env, ssl_);
+  auto name = ssl_.getCipherName();
+  if (!name.has_value()) return Undefined(env->isolate());
+  return OneByteString(env->isolate(), name.value());
 }
 
 MaybeLocal<Value> TLSSession::cipher_version(Environment* env) const {
-  return crypto::GetCurrentCipherVersion(env, ssl_);
+  auto version = ssl_.getCipherVersion();
+  if (!version.has_value()) return Undefined(env->isolate());
+  return OneByteString(env->isolate(), version.value());
 }
 
 const std::string_view TLSSession::servername() const {
-  const char* servername = crypto::GetServerName(ssl_.get());
-  return servername != nullptr ? std::string_view(servername)
-                               : std::string_view();
+  return ssl_.getServerName().value_or(std::string_view());
 }
 
-const std::string_view TLSSession::alpn() const {
+const std::string_view TLSSession::protocol() const {
   const unsigned char* alpn_buf = nullptr;
   unsigned int alpnlen;
   SSL_get0_alpn_selected(ssl_.get(), &alpn_buf, &alpnlen);
@@ -629,7 +645,7 @@ const std::string_view TLSSession::alpn() const {
 }
 
 bool TLSSession::InitiateKeyUpdate() {
-  if (session_->is_destroyed() || in_key_update_) return false;
+  if (in_key_update_) return false;
   auto leave = OnScopeLeave([this] { in_key_update_ = false; });
   in_key_update_ = true;
 
